from trailblazer.apps.slurm.api import cancel_slurm_job, get_slurm_queue
from trailblazer.apps.slurm.models import SqueueResult


class SlurmCLIClient:
    def __init__(self, host: str):
        self.host = host

    def get_slurm_queue(self, job_ids: list[int]) -> SqueueResult:
        return get_slurm_queue(job_ids=job_ids, analysis_host=self.host)

    def cancel_job(self, job_id: int) -> None:
        cancel_slurm_job(slurm_id=job_id, analysis_host=self.host)



from trailblazer.clients.slurm_api_client.dto.job_response import SlurmJobResponse
from trailblazer.clients.tower.models import TowerTasksResponse, TowerWorkflowResponse
from trailblazer.constants import TrailblazerStatus
from trailblazer.services.analysis_service.analysis_service import AnalysisService
from trailblazer.store.models import Analysis


def test_updating_tower_analysis(
    analysis_service: AnalysisService,
    tower_analysis: Analysis,
    tower_tasks_response: TowerTasksResponse,
    tower_workflow_response: TowerWorkflowResponse,
):

    # GIVEN an analysis started with tower without any job entries but running tasks
    analysis_service.job_service.tower_service.client.get_tasks.return_value = tower_tasks_response
    analysis_service.job_service.tower_service.client.get_workflow.return_value = (
        tower_workflow_response
    )

    # WHEN updating the tower analysis
    analysis_service.update_analysis_meta_data(tower_analysis.id)

    # THEN job entries have been created for the analysis
    assert tower_analysis.jobs

    # THEN the status of the analysis is running
    assert tower_analysis.status == TrailblazerStatus.RUNNING

    # THEN the progress of the analysis is updated
    assert tower_analysis.progress > 0


def test_updating_slurm_analysis(
    analysis_service: AnalysisService,
    slurm_analysis: Analysis,
    slurm_job_response: SlurmJobResponse,
):

    # GIVEN an analysis started with slurm without any job entries but running jobs
    analysis_service.job_service.slurm_service.client.get_job.return_value = slurm_job_response

    # WHEN updating the slurm analysis
    analysis_service.update_analysis_meta_data(slurm_analysis.id)

    # THEN job entries have been created for the analysis
    assert slurm_analysis.jobs

    # THEN the status of the analysis is running
    assert slurm_analysis.status == TrailblazerStatus.COMPLETED

    # THEN the progress of the analysis is updated
    assert slurm_analysis.progress > 0



from pydantic import BaseModel

from trailblazer.clients.slurm_api_client.dto.common import (
    SlurmError,
    SlurmAPIJobInfo,
    SlurmWarning,
)


class SlurmJobResponse(BaseModel):
    jobs: list[SlurmAPIJobInfo]
    errors: list[SlurmError] | None = None
    warnings: list[SlurmWarning] | None = None



from datetime import date, datetime

from pydantic import BaseModel

from trailblazer.constants import TrailblazerStatus, WorkflowManager


class User(BaseModel):
    created_at: datetime | None = None
    email: str | None = None
    id: int
    is_archived: bool | None = False
    name: str | None = None
    abbreviation: str | None = None


class Job(BaseModel):
    analysis_id: int
    context: str | None = None
    elapsed: int | None = None
    id: int
    name: str | None = None
    slurm_id: int
    started_at: datetime | None = None
    status: str


class AnalysisResponse(BaseModel):
    id: int
    case_id: str
    comment: str | None = None
    completed_at: datetime | None = None
    config_path: str | None = None
    delivered_by: str | None = None
    delivered_date: date | None = None
    is_visible: bool = True
    is_delivered: bool
    jobs: list[Job] = []
    latest_failed_job: Job | None = None
    logged_at: datetime | None = None
    order_id: int | None = None
    out_dir: str | None = None
    priority: str
    progress: float
    started_at: datetime | None = None
    status: TrailblazerStatus
    ticket_id: str | None = None
    type: str | None = None
    upload_jobs: list[Job] = []
    uploaded_at: datetime | None = None
    user_id: int | None = None
    user: User | None = None
    version: str | None = None
    workflow: str | None = None
    workflow_manager: WorkflowManager



from datetime import datetime, timedelta
from pathlib import Path
from typing import Dict, Generator

import pytest
from sqlalchemy.orm import Session

from tests.mocks.store_mock import MockStore
from tests.store.utils.store_helper import StoreHelpers
from trailblazer.clients.slurm_api_client.dto.common import SlurmAPIJobInfo
from trailblazer.clients.slurm_api_client.dto.job_response import SlurmJobResponse
from trailblazer.clients.tower.models import (
    TaskWrapper,
    TowerProgress,
    TowerTask,
    TowerTasksResponse,
    TowerWorkflow,
    TowerWorkflowResponse,
)
from trailblazer.constants import (
    PRIORITY_OPTIONS,
    TOWER_TIMESTAMP_FORMAT,
    TYPES,
    FileFormat,
    JobType,
    TrailblazerStatus,
    WorkflowManager,
)
from trailblazer.io.controller import ReadFile
from trailblazer.store.database import (
    create_all_tables,
    drop_all_tables,
    get_session,
    initialize_database,
)
from trailblazer.store.models import Analysis, Job
from trailblazer.store.store import Store


class CaseId:
    PENDING: str = "cuddlyhen_pending"
    RUNNING: str = "cuddlyhen"
    COMPLETED: str = "cuddlyhen_completed"


@pytest.fixture(scope="session")
def analysis_id_does_not_exist() -> int:
    return 123456789666


@pytest.fixture(scope="session")
def analysis_comment() -> str:
    """Return a comment."""
    return "a comment"


@pytest.fixture(scope="session")
def username() -> str:
    """Return a username."""
    return "Paul Anderson"


@pytest.fixture(scope="session")
def archived_username() -> str:
    """Return an archived username."""
    return "Archived User"


@pytest.fixture(scope="session")
def archived_user_email() -> str:
    """Return an archived user email."""
    return "archived.user@magnolia.com"


@pytest.fixture(scope="session")
def user_email() -> str:
    """Return an user email."""
    return "paul.anderson@magnolia.com"


@pytest.fixture(scope="session")
def fixtures_dir() -> Path:
    """Return the path to the fixtures' dir."""
    return Path("tests", "fixtures")


@pytest.fixture(scope="session")
def analysis_data_path(fixtures_dir: Path) -> Path:
    """Return the path to an analysis data file."""
    return Path(fixtures_dir, "analysis-data.yaml")


@pytest.fixture(scope="function")
def analysis_data(analysis_data_path: Path) -> dict[str, list]:
    """Return content of the analysis data file."""
    return ReadFile.get_content_from_file(file_format=FileFormat.YAML, file_path=analysis_data_path)


@pytest.fixture(scope="session")
def squeue_stream_jobs() -> str:
    """Return a squeue output stream."""
    return """JOBID,NAME,STATE,TIME_LIMIT,TIME,START_TIME
690994,gatk_genotypegvcfs2,COMPLETED,10:00:00,1:01:52,2020-10-22T11:43:33
702463,bwa_mem_ACC3218A1,COMPLETED,1-06:00:00,1-1:28:12,2020-10-27T23:06:34
690992,gatk_genotypegvcfs3,COMPLETED,10:00:00,5:54,2020-10-22T11:43:02
690988,gatk_genotypegvcfs4,RUNNING,10:00:00,0:19,N/A
690989,gatk_genotypegvcfs5,PENDING,10:00:00,0:00,N/A"""


@pytest.fixture
def trailblazer_context(analysis_store: MockStore) -> dict[str, MockStore]:
    """Trailblazer context to be used in CLI."""
    return {"trailblazer_db": analysis_store}


@pytest.fixture
def store() -> Generator[MockStore, None, None]:
    """Empty Trailblazer database."""
    initialize_database("sqlite:///:memory:")
    _store = Store()
    create_all_tables()
    yield _store
    drop_all_tables()


@pytest.fixture(scope="function")
def info_store(store: MockStore) -> Generator[MockStore, None, None]:
    """A Trailblazer database wih a populated info table."""
    StoreHelpers.add_info(store=store)
    yield store


@pytest.fixture(scope="function")
def job_store(
    store: MockStore,
) -> Generator[MockStore, None, None]:
    """A Trailblazer database wih a populated job table."""
    statuses: list[str] = [TrailblazerStatus.COMPLETED, TrailblazerStatus.FAILED]
    for index, status in enumerate(statuses):
        StoreHelpers.add_job(analysis_id=index, name=str(index), slurm_id=index, status=status)
    yield store


@pytest.fixture(scope="function")
def user_store(
    archived_user_email: str,
    archived_username: str,
    user_email: str,
    username: str,
    store: MockStore,
) -> Generator[MockStore, None, None]:
    """A Trailblazer database wih a populated user table."""
    StoreHelpers.add_user(email=user_email, name=username, abbreviation="SUS")
    StoreHelpers.add_user(
        email=archived_user_email,
        name=archived_username,
        abbreviation="SUSUS",
        is_archived=True,
    )
    yield store


@pytest.fixture(scope="function")
def raw_analyses(analysis_data: dict[str, list[Dict]]) -> list[dict]:
    """Return raw analyses data."""
    analyses: list[dict] = []
    for analysis in analysis_data["analyses"]:
        analyses.append(analysis)
    return analyses


@pytest.fixture
def analysis_store(
    analysis_data: dict[str, list],
    archived_user_email: str,
    archived_username: str,
    raw_analyses: list[dict],
    store: MockStore,
) -> Generator[MockStore, None, None]:
    """A sample Trailblazer database populated with pending analyses."""
    session: Session = get_session()
    StoreHelpers.add_user(
        email=archived_user_email, name=archived_username, is_archived=True, abbreviation="abbrev"
    )
    for user_data in analysis_data["users"]:
        store.add_user(
            name=user_data["name"],
            email=user_data["email"],
            abbreviation=user_data["abbreviation"],
        )
    for raw_analysis in raw_analyses:
        raw_analysis["user"] = store.get_user(email=raw_analysis["user"])
        raw_analysis["case_id"] = raw_analysis.pop("case_id")
        session.add(Analysis(**raw_analysis))
    yield store


@pytest.fixture(scope="session")
def timestamp_now() -> datetime:
    """Return a time stamp of today's date in date time format."""
    return datetime.now()


@pytest.fixture(scope="session")
def timestamp_yesterday(timestamp_now: datetime) -> datetime:
    """Return a time stamp of yesterday's date in date time format."""
    return timestamp_now - timedelta(days=1)


@pytest.fixture(scope="session")
def timestamp_old() -> datetime:
    """Return an old time stamp in a date time format."""
    timestamp: str = "1066-10-14 00:00:00"
    timestamp_format: str = "%Y-%m-%d %H:%M:%S"
    return datetime.strptime(timestamp, timestamp_format)


@pytest.fixture
def tower_jobs(analysis_id, started_at, slurm_id, tower_task_name) -> list[dict]:
    """Return a list of Tower Jobs."""
    return [
        dict(
            analysis_id=analysis_id,
            slurm_id=slurm_id,
            name=tower_task_name,
            started_at=started_at,
            elapsed=63,
            status=TrailblazerStatus.COMPLETED,
        ),
        dict(
            analysis_id=analysis_id,
            slurm_id="4611829",
            name="NFCORE_RNAFUSION:RNAFUSION:PIZZLY_WORKFLOW:KALLISTO_QUANT",
            started_at=None,
            elapsed=0,
            status=TrailblazerStatus.PENDING,
        ),
        dict(
            analysis_id=analysis_id,
            slurm_id="4611828",
            name="NFCORE_RNAFUSION:RNAFUSION:FASTQC",
            started_at=None,
            elapsed=0,
            status=TrailblazerStatus.PENDING,
        ),
    ]


@pytest.fixture
def analysis_id() -> int:
    """Return a mock ID of the analysis in the Trailblazer database."""
    return 1


@pytest.fixture(scope="session")
def started_at() -> datetime:
    """Returns a started at date."""
    return datetime.strptime("2023-04-04T08:11:27Z", TOWER_TIMESTAMP_FORMAT)


@pytest.fixture(scope="session")
def slurm_id() -> str:
    """Returns a slurm id."""
    return "4611827"


@pytest.fixture(scope="session")
def tower_task_name() -> str:
    """Returns a NF Tower task name."""
    return "NFCORE_RNAFUSION:RNAFUSION:INPUT_CHECK:SAMPLESHEET_CHECK"


@pytest.fixture(scope="session")
def case_id() -> str:
    """Return a case id."""
    return CaseId.RUNNING


@pytest.fixture(scope="session")
def case_id_not_in_db() -> str:
    """Return a case id not present in database."""
    return "case_id_not_in_db"


@pytest.fixture(scope="session")
def failed_analysis_case_id() -> str:
    """Return a case id for a failed analysis."""
    return "crackpanda"


@pytest.fixture
def analysis_with_upload_jobs(case_id: str) -> Analysis:
    upload_job = Job(
        name="upload",
        slurm_id="123456",
        job_type=JobType.UPLOAD,
        status=TrailblazerStatus.COMPLETED,
        started_at=datetime.now(),
        elapsed=10,
    )
    analysis = Analysis(case_id=case_id)
    analysis.jobs.append(upload_job)
    session = get_session()
    session.add(analysis)
    return analysis


@pytest.fixture
def tower_analysis(analysis_store: Store, tmp_path) -> Analysis:
    case_id = "case_id"
    config_path: Path = tmp_path / "tower.yaml"
    config_path.write_text(f"case_id: {case_id}")

    analysis = Analysis(
        config_path=str(config_path),
        workflow="workflow",
        case_id="case_id",
        out_dir="out_dir",
        priority=PRIORITY_OPTIONS[0],
        started_at=datetime.now() - timedelta(weeks=1),
        status=TrailblazerStatus.PENDING,
        ticket_id="ticket_id",
        type=TYPES[0],
        workflow_manager=WorkflowManager.TOWER,
        is_visible=True,
        order_id=1,
    )
    session: Session = get_session()
    session.add(analysis)
    session.commit()
    return analysis


@pytest.fixture
def tower_tasks_response():
    task_1 = TowerTask(
        process="example_process",
        name="example_task",
        status="COMPLETED",
        nativeId="1234",
        dateCreated=datetime.now(),
        lastUpdated=datetime.now(),
        start=datetime.now(),
        module=["example_module"],
    )

    task_2 = TowerTask(
        process="example_process",
        name="example_task",
        status="RUNNING",
        nativeId="1234",
        dateCreated=datetime.now(),
        lastUpdated=datetime.now(),
        start=datetime.now(),
        module=["example_module"],
    )

    task_wrapper_1 = TaskWrapper(task=task_1)
    task_wrapper_2 = TaskWrapper(task=task_2)
    return TowerTasksResponse(tasks=[task_wrapper_1, task_wrapper_2], total=2)


@pytest.fixture
def tower_workflow_response() -> TowerWorkflowResponse:
    worfklow = TowerWorkflow(status="RUNNING")
    progress = TowerProgress(workflowProgress={}, processesProgress=[])
    return TowerWorkflowResponse(
        workflow=worfklow,
        progress=progress,
    )


@pytest.fixture
def slurm_analysis(tmp_path) -> Analysis:
    config_path: Path = tmp_path / "slurm.yaml"
    content = """---
    case_id:
    - '6123780'
    """
    config_path.write_text(content)

    analysis = Analysis(
        config_path=str(config_path),
        workflow="workflow",
        case_id="case_id",
        out_dir="out_dir",
        priority=PRIORITY_OPTIONS[0],
        started_at=datetime.now() - timedelta(weeks=1),
        status=TrailblazerStatus.PENDING,
        ticket_id="ticket_id",
        type=TYPES[0],
        workflow_manager=WorkflowManager.SLURM,
        is_visible=True,
        order_id=1,
    )
    session: Session = get_session()
    session.add(analysis)
    session.commit()
    return analysis


@pytest.fixture
def slurm_job_response() -> SlurmJobResponse:
    job_info = SlurmAPIJobInfo(
        job_id=12345,
        job_state=["COMPLETED"],
        name="Test Job",
    )
    return SlurmJobResponse(jobs=[job_info], errors=None, warnings=None)



from datetime import datetime, timedelta
from unittest.mock import MagicMock, Mock
import pytest
from sqlalchemy.orm import Session


from trailblazer.clients.tower.tower_client import TowerAPIClient
from trailblazer.constants import (
    PRIORITY_OPTIONS,
    TYPES,
    JobType,
    TrailblazerStatus,
    WorkflowManager,
)
from trailblazer.services.analysis_service.analysis_service import AnalysisService
from trailblazer.services.job_service.job_service import JobService
from trailblazer.services.slurm.dtos import SlurmJobInfo
from trailblazer.services.slurm.slurm_cli_service.slurm_cli_service import SlurmCLIService
from trailblazer.services.slurm.slurm_service import SlurmService
from trailblazer.services.tower.tower_api_service import TowerAPIService
from trailblazer.store.database import get_session
from trailblazer.store.models import Analysis, Job
from trailblazer.store.store import Store


@pytest.fixture
def slurm_completed_job_info() -> SlurmJobInfo:
    return SlurmJobInfo(
        slurm_id=690994,
        name="job",
        status="completed",
        elapsed=1000,
        started_at="2020-10-22T11:43:33",
    )


@pytest.fixture
def tower_service(store: Store) -> TowerAPIService:
    tower_client = MagicMock(spec=TowerAPIClient)
    return TowerAPIService(client=tower_client, store=store)


@pytest.fixture
def slurm_service(slurm_completed_job_info) -> SlurmService:
    """Slurm service reporting all jobs as completed."""
    service = SlurmCLIService(client=Mock(), store=Mock())
    service.get_job = Mock(return_value=slurm_completed_job_info)
    service.update_jobs = Mock(return_value=[slurm_completed_job_info])
    return service


@pytest.fixture
def job_service(
    analysis_store: Store,
    slurm_service: SlurmService,
    tower_service: TowerAPIService,
) -> JobService:
    return JobService(
        store=analysis_store,
        slurm_service=slurm_service,
        tower_service=tower_service,
    )


@pytest.fixture
def job_service_mock():
    return MagicMock(spec=JobService)


@pytest.fixture
def analysis(analysis_store: Store) -> Analysis:
    analysis = Analysis(
        config_path="config_path",
        workflow="workflow",
        case_id="case_id",
        out_dir="out_dir",
        priority=PRIORITY_OPTIONS[0],
        started_at=datetime.now() - timedelta(weeks=1),
        status=TrailblazerStatus.RUNNING,
        ticket_id="ticket_id",
        type=TYPES[0],
        workflow_manager=WorkflowManager.SLURM,
        is_visible=True,
        order_id=1,
    )
    session: Session = get_session()
    session.add(analysis)
    session.commit()
    return analysis


@pytest.fixture
def analysis_with_running_jobs(analysis_store: Store, analysis: Analysis) -> Analysis:
    analysis_job_1 = Job(
        analysis_id=analysis.id,
        name="name",
        slurm_id=1,
        status=TrailblazerStatus.RUNNING,
        started_at=datetime.now(),
        elapsed=100,
        job_type=JobType.ANALYSIS,
    )
    analysis_job_2 = Job(
        analysis_id=analysis.id,
        name="name",
        slurm_id=2,
        status=TrailblazerStatus.COMPLETED,
        started_at=datetime.now(),
        elapsed=1,
        job_type=JobType.ANALYSIS,
    )
    session: Session = get_session()
    session.add_all([analysis_job_1, analysis_job_2])
    session.commit()
    return analysis


@pytest.fixture
def analysis_with_completed_jobs(analysis_store: Store, analysis: Analysis) -> Analysis:
    analysis_job_1 = Job(
        analysis_id=analysis.id,
        name="name",
        slurm_id=1,
        status=TrailblazerStatus.COMPLETED,
        started_at=datetime.now(),
        elapsed=100,
        job_type=JobType.ANALYSIS,
    )
    analysis_job_2 = Job(
        analysis_id=analysis.id,
        name="name",
        slurm_id=2,
        status=TrailblazerStatus.COMPLETED,
        started_at=datetime.now(),
        elapsed=1,
        job_type=JobType.ANALYSIS,
    )
    session: Session = get_session()
    session.add_all([analysis_job_1, analysis_job_2])
    session.commit()
    return analysis


@pytest.fixture
def analysis_without_jobs(analysis_store: Store):
    analysis = Analysis(
        config_path="config_path",
        workflow="workflow",
        case_id="case_id",
        out_dir="out_dir",
        priority=PRIORITY_OPTIONS[0],
        started_at=datetime.now() - timedelta(weeks=1),
        status=TrailblazerStatus.RUNNING,
        ticket_id="ticket_id",
        type=TYPES[0],
        workflow_manager=WorkflowManager.SLURM,
        is_visible=True,
        order_id=1,
    )
    session: Session = get_session()
    session.add(analysis)
    session.commit()
    return analysis


@pytest.fixture
def analysis_service(analysis_store: Store, job_service_mock: JobService) -> AnalysisService:
    return AnalysisService(store=analysis_store, job_service=job_service_mock)



from pydantic import BaseModel


class SlurmError(BaseModel):
    description: str | None = None
    error_number: int | None = None
    error: str | None = None
    source: str | None = None


class SlurmWarning(BaseModel):
    description: str | None = None
    source: str | None = None


class NumberWithFlags(BaseModel):
    set: bool | None = None
    infinite: bool | None = None
    number: int | None = None


class SlurmAPIJobInfo(BaseModel):
    accrue_time: NumberWithFlags | None = None
    job_id: int | None = None
    job_state: list[str] | None = None
    name: str | None = None
    start_time: NumberWithFlags | None = None



class TrailblazerError(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return f"TrailblazerError: {self.message}"


class MissingAnalysis(TrailblazerError):
    """Error for missing analysis"""


class MissingJob(TrailblazerError):
    """Error for missing job"""


class CancelSlurmAnalysisNotSupportedError(TrailblazerError):
    def __init__(self):
        self.message = "Cancelling SLURM analysis via the web app is not supported"


class MissingFileError(TrailblazerError):
    pass


class EmptySqueueError(TrailblazerError):
    pass


class TowerError(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return f"TowerError: {self.message}"


class MissingSqueueOutput(TrailblazerError):
    """Raised when squeue results sre missing."""


class TowerRequirementsError(TowerError):
    """Raised when Tower requirements are not satisfied."""


class TowerResponseError(TowerError):
    """Raised when an invalid Tower response is returned."""


class ValidationError(TrailblazerError):
    """
    Exception related to validations.
    """


class SlurmAPIClientError(TrailblazerError):
    """
    Exception related to Slurm API client.
    """


class ResponseDeserializationError(SlurmAPIClientError):
    """
    Exception related to the data in a  Slurm API client response.
    """


class TowerAPIClientError(TrailblazerError):
    """
    Exception related to Tower API client.
    """


class InvalidTowerAPIResponse(TowerAPIClientError):
    """
    Exception raised when the Tower API response is not formatted as expected.
    """


class TowerRequestFailed(TowerAPIClientError):
    """
    Exception raised when the Tower API request fails.
    """



from dataclasses import dataclass
from typing import Type

from sqlalchemy import func
from sqlalchemy.orm import Query, Session

from trailblazer.store.database import get_session
from trailblazer.store.models import Job, Model


@dataclass
class BaseHandler:
    """All models in one base class."""

    def get_query(self, table: Type[Model]) -> Query:
        """Return a query for the given table."""
        session: Session = get_session()
        return session.query(table)

    def get_job_query_with_name_and_count_labels(self) -> Query:
        """Return a Job query with a name label and a count with a label."""
        session: Session = get_session()
        return session.query(
            Job.name.label("name"),
            func.count(Job.id).label("count"),
        )



from tests.mocks.store_mock import MockStore
from trailblazer.store.filters.user_filters import UserFilter, apply_user_filter
from trailblazer.store.models import User


def test_add_user(store: MockStore, user_email: str, username: str):
    """Test adding a user to the database."""
    # GIVEN an empty database
    assert not store.get_query(table=User).first()

    # WHEN adding a new user
    new_user: User = store.add_user(name=username, email=user_email, abbreviation="abb")

    # THEN it should be stored in the database
    user: User = apply_user_filter(
        filter_functions=[UserFilter.BY_EMAIL],
        users=store.get_query(table=User),
        email=user_email,
    ).first()
    assert new_user
    assert user == new_user



from abc import ABC, abstractmethod

from trailblazer.services.slurm.dtos import SlurmJobInfo


class SlurmService(ABC):
    @abstractmethod
    def get_job(self, job_id: int) -> SlurmJobInfo:
        pass

    @abstractmethod
    def update_jobs(self, analysis_id: int) -> None:
        pass

    @abstractmethod
    def cancel_jobs(self, analysis_id: int) -> None:
        pass



from datetime import datetime, timedelta
from pathlib import Path
import pytest
from sqlalchemy.orm import Session

from trailblazer.clients.tower.models import TaskWrapper, TowerTask, TowerTasksResponse
from trailblazer.constants import (
    PRIORITY_OPTIONS,
    TYPES,
    TrailblazerStatus,
    WorkflowManager,
)
from trailblazer.store.database import get_session
from trailblazer.store.models import Analysis
from trailblazer.store.store import Store


@pytest.fixture
def tower_analysis(analysis_store: Store, tmp_path) -> Analysis:
    case_id = "case_id"
    config_path: Path = tmp_path / "tower.yaml"
    config_path.write_text(f"case_id: {case_id}")

    analysis = Analysis(
        config_path=str(config_path),
        workflow="workflow",
        case_id="case_id",
        out_dir="out_dir",
        priority=PRIORITY_OPTIONS[0],
        started_at=datetime.now() - timedelta(weeks=1),
        status=TrailblazerStatus.RUNNING,
        ticket_id="ticket_id",
        type=TYPES[0],
        workflow_manager=WorkflowManager.TOWER,
        is_visible=True,
        order_id=1,
    )
    session: Session = get_session()
    session.add(analysis)
    session.commit()
    return analysis


@pytest.fixture
def tower_tasks_response():
    task = TowerTask(
        process="example_process",
        name="example_task",
        status="COMPLETED",
        nativeId="1234",
        dateCreated=datetime.now(),
        lastUpdated=datetime.now(),
        start=datetime.now(),
        module=["example_module"],
    )

    task_wrapper = TaskWrapper(task=task)
    return TowerTasksResponse(tasks=[task_wrapper], total=1)



from trailblazer.services.encryption_service.utils import decrypt_data, encrypt_data


class EncryptionService:
    def __init__(self, secret_key: str):
        self.secret_key = secret_key

    def encrypt(self, data: str) -> str:
        return encrypt_data(data=data, secret_key=self.secret_key)

    def decrypt(self, data: str) -> str:
        return decrypt_data(data=data, secret_key=self.secret_key)



from trailblazer.services.encryption_service.encryption_service import EncryptionService


def test_encryption(encryption_service: EncryptionService):
    # GIVEN a plain text
    plain_text = "my secret"

    # WHEN encrypting and decrypting the plain text
    encrypted_text: str = encryption_service.encrypt(plain_text)
    decrypted_text: str = encryption_service.decrypt(encrypted_text)

    # THEN the decrypted text is the same as the plain text
    assert decrypted_text == plain_text

    # THEN the encrypted text is different from the plain text
    assert encrypted_text != plain_text



import requests

from trailblazer.clients.authentication_client.dtos.refresh_token_request import (
    RefreshAccessTokenRequest,
)
from trailblazer.clients.authentication_client.dtos.tokens_request import GetTokensRequest
from trailblazer.clients.authentication_client.dtos.tokens_response import TokensResponse
from trailblazer.clients.authentication_client.exceptions import GoogleOAuthClientError


class GoogleOAuthClient:
    def __init__(self, client_id: str, client_secret: str, redirect_uri: str, oauth_base_url: str):
        self.client_id = client_id
        self.client_secret = client_secret
        self.oauth_base_url = oauth_base_url
        self.redirect_uri = redirect_uri

    def get_tokens(self, authorization_code: str) -> TokensResponse:
        """Exchange the authorization code for an access token and refresh token."""
        request = GetTokensRequest(
            client_id=self.client_id,
            client_secret=self.client_secret,
            code=authorization_code,
            redirect_uri=self.redirect_uri,
        )
        data: dict = request.model_dump()

        response = requests.post(self.oauth_base_url, data=data)

        if not response.ok:
            raise GoogleOAuthClientError(response.text)

        return TokensResponse.model_validate(response.json())

    def get_id_token(self, refresh_token: str) -> str:
        """Use refresh token to get a new id token."""
        request = RefreshAccessTokenRequest(
            client_id=self.client_id,
            client_secret=self.client_secret,
            refresh_token=refresh_token,
        )
        data: str = request.model_dump_json()

        response = requests.post(self.oauth_base_url, data=data)

        if not response.ok:
            raise GoogleOAuthClientError(response.text)

        return response.json()["id_token"]



from requests_mock import Mocker

from trailblazer.services.authentication_service.authentication_service import AuthenticationService
from trailblazer.services.encryption_service.encryption_service import EncryptionService
from trailblazer.store.models import User


def test_exchange_code(authentication_service: AuthenticationService, user_email: str):
    # GIVEN an authentication service

    # WHEN exchanging the authorization code
    token: str = authentication_service.authenticate("auth_code")

    # THEN an access token is returned
    assert token

    # THEN the refresh token is stored on the user
    user: User = authentication_service.store.get_user(user_email)
    assert user.refresh_token


def test_refresh_access_token(
    authentication_service: AuthenticationService,
    encryption_service: EncryptionService,
    user_email: str,
):
    # GIVEN an encrypted refresh token
    encrypted_refresh_token: str = encryption_service.encrypt("refresh_token")

    # GIVEN an existing user with the refresh token
    user: User = authentication_service.store.get_user(user_email)
    user.refresh_token = encrypted_refresh_token

    # WHEN refreshing the token
    token: str = authentication_service.refresh_token(user.id)

    # THEN a token is returned
    assert token



from pydantic import BaseModel

from trailblazer.clients.slurm_api_client.dto.common import SlurmError, SlurmWarning


class SlurmCancelJobResponse(BaseModel):
    errors: list[SlurmError] | None
    warnings: list[SlurmWarning] | None



import logging
from datetime import datetime
from sqlalchemy.orm import Session

from trailblazer.constants import TrailblazerStatus
from trailblazer.dto.update_analyses import UpdateAnalyses
from trailblazer.exc import MissingAnalysis
from trailblazer.services.slurm.dtos import SlurmJobInfo
from trailblazer.store.base import BaseHandler
from trailblazer.store.database import get_session
from trailblazer.store.models import Analysis, Delivery, Job, User

LOG = logging.getLogger(__name__)


class UpdateHandler(BaseHandler):
    """Class for updating items in the database."""

    def update_jobs(self, jobs: list[dict]) -> None:
        """Update jobs in the analysis."""
        session: Session = get_session()
        for job in jobs:
            updated_job = Job(**job)
            session.add(updated_job)
        session.commit()

    def update_user_is_archived(self, user: User, archive: bool = True) -> None:
        """Update is archived for a user in the database."""
        user.is_archived = archive
        session: Session = get_session()
        session.commit()

    def get_ongoing_analyses(self) -> list[Analysis]:
        """Return all analyses with ongoing status."""
        ongoing_statuses: list[str] = list(TrailblazerStatus.ongoing_statuses())
        return self.get_analyses_with_statuses(ongoing_statuses)

    def update_analysis_status_by_case_id(self, case_id: str, status: str):
        """Setting analysis status."""
        status: str = status.lower()
        if status not in set(TrailblazerStatus.statuses()):
            raise ValueError(f"Invalid status. Allowed values are: {TrailblazerStatus.statuses()}")
        analysis: Analysis | None = self.get_latest_analysis_for_case(case_id)
        analysis.status = status
        session: Session = get_session()
        session.commit()
        LOG.info(f"{analysis.case_id} - Status set to {status.upper()}")

    def update_analysis_status_to_completed(self, analysis_id: int) -> None:
        """Set an analysis status to 'completed'."""
        analysis: Analysis | None = self.get_analysis_with_id(analysis_id)
        self.update_analysis_status_by_case_id(
            case_id=analysis.case_id, status=TrailblazerStatus.COMPLETED
        )

    def update_analysis_uploaded_at(self, case_id: str, uploaded_at: datetime) -> None:
        """Set analysis uploaded at for an analysis."""
        analysis: Analysis | None = self.get_latest_analysis_for_case(case_id)
        analysis.uploaded_at = uploaded_at
        session: Session = get_session()
        session.commit()

    def update_latest_analysis_comment(self, case_id: str, comment: str) -> None:
        analysis: Analysis | None = self.get_latest_analysis_for_case(case_id)
        self.update_analysis_comment(analysis_id=analysis.id, comment=comment)

    def update_analysis_comment(self, analysis_id: int, comment: str):
        analysis: Analysis | None = self.get_analysis_with_id(analysis_id)
        analysis.comment = " ".join([analysis.comment, comment]) if analysis.comment else comment
        session: Session = get_session()
        session.commit()
        LOG.info(f"Adding comment {comment} to analysis {analysis.case_id}")

    @staticmethod
    def update_analysis_delivery(
        analysis: Analysis, is_delivered: bool, user: User | None = None
    ) -> None:
        session: Session = get_session()
        if is_delivered and not analysis.delivery:
            delivery = Delivery(
                analysis_id=analysis.id,
                delivered_by=user.id,
                delivered_date=datetime.today(),
            )
            session.add(delivery)
            session.commit()
            analysis.delivery = delivery
        elif not is_delivered:
            if delivery := analysis.delivery:
                session.delete(delivery)

    def update_analysis(
        self,
        analysis_id: int,
        comment: str | None = None,
        is_delivered: bool | None = None,
        is_visible: bool | None = None,
        status: str | None = None,
        user: User | None = None,
    ) -> Analysis:
        """Update an analysis."""
        analysis: Analysis | None = self.get_analysis_with_id(analysis_id)

        if not analysis:
            raise MissingAnalysis(f"Analysis {analysis_id} does not exist")

        if comment is not None:
            LOG.info(f"Adding comment {comment} to analysis {analysis.id}")
            analysis.comment = comment

        if is_delivered is not None:
            LOG.info(f"Setting analysis delivered status to {is_delivered}")
            self.update_analysis_delivery(analysis=analysis, is_delivered=is_delivered, user=user)

        if is_visible is not None:
            LOG.info(f"Setting visibility to {is_visible} for analysis {analysis.id}")
            analysis.is_visible = bool(is_visible)

        if status is not None:
            LOG.info(f"Setting status to {status} for analysis {analysis.id}")
            analysis.status = status

        session: Session = get_session()
        session.commit()

        return analysis

    def update_analyses(self, data: UpdateAnalyses, user: User | None = None) -> list[Analysis]:
        updated_analyses: list[Analysis] = []
        for analysis_update in data.analyses:
            analysis: Analysis = self.update_analysis(
                analysis_id=analysis_update.id,
                is_delivered=analysis_update.is_delivered,
                comment=analysis_update.comment,
                is_visible=analysis_update.is_visible,
                status=analysis_update.status,
                user=user,
            )
            updated_analyses.append(analysis)
        return updated_analyses

    def update_job(self, job_id: int, job_info: SlurmJobInfo) -> Job:
        job: Job | None = self.get_job_by_id(job_id)
        job.name = job_info.name
        job.status = job_info.status
        job.elapsed = job_info.elapsed
        job.started_at = job_info.started_at
        session: Session = get_session()
        session.commit()

    def update_user_token(self, refresh_token: str, user_id: int) -> None:
        user: User | None = self.get_user_by_id(user_id)
        user.refresh_token = refresh_token
        session: Session = get_session()
        session.commit()

    def update_analysis_status(self, analysis_id: int, status: TrailblazerStatus) -> None:
        analysis: Analysis | None = self.get_analysis_with_id(analysis_id)
        analysis.status = status
        session: Session = get_session()
        session.commit()
        LOG.info(f"Updated status {analysis.case_id} - {analysis.id}: {analysis.status} ")

    def update_analysis_progress(self, analysis_id: int, progress: float) -> None:
        analysis: Analysis | None = self.get_analysis_with_id(analysis_id)
        analysis.progress = progress
        session: Session = get_session()
        session.commit()

    def update_analysis_upload_date(self, analysis_id: int, uploaded_at: datetime) -> None:
        analysis: Analysis | None = self.get_analysis_with_id(analysis_id)
        analysis.uploaded_at = uploaded_at
        session: Session = get_session()
        session.commit()



import os
from dependency_injector import containers, providers

from trailblazer.clients.authentication_client.google_oauth_client import GoogleOAuthClient
from trailblazer.clients.google_api_client.google_api_client import GoogleAPIClient
from trailblazer.clients.slurm_api_client.slurm_api_client import SlurmAPIClient
from trailblazer.clients.slurm_cli_client.slurm_cli_client import SlurmCLIClient
from trailblazer.clients.tower.tower_client import TowerAPIClient
from trailblazer.services.analysis_service.analysis_service import AnalysisService
from trailblazer.services.authentication_service.authentication_service import AuthenticationService
from trailblazer.services.encryption_service.encryption_service import EncryptionService
from trailblazer.services.job_service import JobService
from trailblazer.services.slurm.slurm_api_service.slurm_api_service import SlurmAPIService
from trailblazer.services.slurm.slurm_cli_service.slurm_cli_service import SlurmCLIService
from trailblazer.services.tower.tower_api_service import TowerAPIService
from trailblazer.store.store import Store


class Container(containers.DeclarativeContainer):
    slurm_host: str | None = os.environ.get("ANALYSIS_HOST")
    oauth_client_id: str | None = os.environ.get("GOOGLE_CLIENT_ID")
    oauth_client_secret: str | None = os.environ.get("GOOGLE_CLIENT_SECRET")
    oauth_redirect_uri: str | None = os.environ.get("GOOGLE_REDIRECT_URI")
    google_oauth_base_url: str | None = os.environ.get("GOOGLE_OAUTH_BASE_URL")
    encryption_key: str | None = os.environ.get("ENCRYPTION_KEY")
    google_api_base_url: str | None = os.environ.get("GOOGLE_API_BASE_URL")
    slurm_jwt_token: str | None = os.environ.get("SLURM_JWT")
    slurm_user_name: str | None = os.environ.get("SLURM_USER_NAME")
    slurm_base_url: str | None = os.environ.get("SLURM_BASE_URL")
    tower_base_url: str | None = os.environ.get("TOWER_API_ENDPOINT")
    tower_access_token: str | None = os.environ.get("TOWER_ACCESS_TOKEN")
    tower_workspace_id: str | None = os.environ.get("TOWER_WORKSPACE_ID")

    google_api_client = GoogleAPIClient(google_api_base_url)

    google_oauth_client = providers.Singleton(
        GoogleOAuthClient,
        client_id=oauth_client_id,
        client_secret=oauth_client_secret,
        redirect_uri=oauth_redirect_uri,
        oauth_base_url=google_oauth_base_url,
    )

    store = providers.Singleton(Store)

    if slurm_base_url and slurm_jwt_token and slurm_user_name:
        slurm_client = providers.Singleton(
            SlurmAPIClient,
            base_url=slurm_base_url,
            access_token=slurm_jwt_token,
            user_name=slurm_user_name,
        )
        slurm_service = providers.Singleton(SlurmAPIService, client=slurm_client)
    else:
        slurm_client = providers.Singleton(SlurmCLIClient, host=slurm_host)
        slurm_service = providers.Singleton(SlurmCLIService, client=slurm_client, store=store)

    tower_client = providers.Singleton(
        TowerAPIClient,
        base_url=tower_base_url,
        access_token=tower_access_token,
        workspace_id=tower_workspace_id,
    )

    tower_service = providers.Singleton(
        TowerAPIService,
        client=tower_client,
        store=store,
    )

    job_service = providers.Factory(
        JobService,
        store=store,
        slurm_service=slurm_service,
        tower_service=tower_service,
    )
    analysis_service = providers.Factory(AnalysisService, store=store, job_service=job_service)

    encryption_service = providers.Singleton(EncryptionService, secret_key=encryption_key)

    auth_service = providers.Singleton(
        AuthenticationService,
        google_oauth_client=google_oauth_client,
        google_api_client=google_api_client,
        encryption_service=encryption_service,
        store=store,
    )



from trailblazer.clients.authentication_client.dtos.tokens_response import TokensResponse
from trailblazer.clients.authentication_client.google_oauth_client import GoogleOAuthClient
from trailblazer.clients.google_api_client.google_api_client import GoogleAPIClient
from trailblazer.services.authentication_service.exceptions import UserNotFoundError
from trailblazer.services.encryption_service.encryption_service import EncryptionService
from trailblazer.store.models import User
from trailblazer.store.store import Store


class AuthenticationService:
    def __init__(
        self,
        google_oauth_client: GoogleOAuthClient,
        google_api_client: GoogleAPIClient,
        encryption_service: EncryptionService,
        store: Store,
    ):
        self.google_oauth_client = google_oauth_client
        self.google_api_client = google_api_client
        self.encryption_service = encryption_service
        self.store = store

    def authenticate(self, authorization_code: str) -> str:
        """Exchange the authorization code for an id token."""
        tokens: TokensResponse = self.google_oauth_client.get_tokens(authorization_code)
        user_email: str = self.google_api_client.get_user_email(tokens.access_token)
        user: User | None = self.store.get_user(user_email)

        if not user:
            raise UserNotFoundError

        encrypted_token: str = self.encryption_service.encrypt(tokens.refresh_token)
        self.store.update_user_token(user_id=user.id, refresh_token=encrypted_token)

        return tokens.id_token

    def refresh_token(self, user_id: int) -> str:
        """Retrieve a refreshed id token for the user."""
        user: User = self.store.get_user_by_id(user_id)
        refresh_token: str = self.encryption_service.decrypt(user.refresh_token)
        id_token: str = self.google_oauth_client.get_id_token(refresh_token)
        return id_token



from unittest.mock import MagicMock
import pytest

from trailblazer.clients.slurm_api_client.slurm_api_client import SlurmAPIClient
from trailblazer.clients.tower.tower_client import TowerAPIClient
from trailblazer.services.analysis_service.analysis_service import AnalysisService
from trailblazer.services.job_service.job_service import JobService
from trailblazer.services.slurm.slurm_api_service.slurm_api_service import SlurmAPIService
from trailblazer.services.tower.tower_api_service import TowerAPIService
from trailblazer.store.store import Store


@pytest.fixture
def analysis_service(analysis_store: Store) -> AnalysisService:
    slurm_client = MagicMock(spec=SlurmAPIClient)
    tower_client = MagicMock(spec=TowerAPIClient)

    slurm_service = SlurmAPIService(store=analysis_store, client=slurm_client)
    tower_service = TowerAPIService(store=analysis_store, client=tower_client)

    job_service = JobService(
        store=analysis_store,
        slurm_service=slurm_service,
        tower_service=tower_service,
    )

    return AnalysisService(
        store=analysis_store,
        job_service=job_service,
    )



import base64
from datetime import datetime, timedelta
import os
import pytest
from requests_mock import Mocker
from sqlalchemy.orm import Session


from trailblazer.clients.authentication_client.google_oauth_client import GoogleOAuthClient
from trailblazer.clients.google_api_client.google_api_client import GoogleAPIClient
from trailblazer.clients.slurm_cli_client.slurm_cli_client import SlurmCLIClient
from trailblazer.clients.tower.tower_client import TowerAPIClient
from trailblazer.constants import (
    PRIORITY_OPTIONS,
    TYPES,
    JobType,
    TrailblazerStatus,
    WorkflowManager,
)
from trailblazer.services.authentication_service.authentication_service import AuthenticationService
from trailblazer.services.encryption_service.encryption_service import EncryptionService
from trailblazer.services.job_service import JobService
from trailblazer.services.slurm.dtos import SlurmJobInfo
from trailblazer.services.slurm.slurm_cli_service.slurm_cli_service import SlurmCLIService

from trailblazer.services.tower.tower_api_service import TowerAPIService
from trailblazer.store.database import get_session
from trailblazer.store.models import Analysis, Job
from trailblazer.store.store import Store


@pytest.fixture
def upload_job_info() -> SlurmJobInfo:
    return SlurmJobInfo(
        slurm_id=123,
        name="upload",
        status=TrailblazerStatus.COMPLETED,
        started_at=datetime.now(),
        elapsed=100,
    )


@pytest.fixture
def tower_service(analysis_store: Store) -> TowerAPIService:
    tower_client = TowerAPIClient(
        base_url="https://tower",
        access_token="token",
        workspace_id="workspace_id",
    )
    return TowerAPIService(client=tower_client, store=analysis_store)


@pytest.fixture
def slurm_service(analysis_store: Store) -> SlurmCLIService:
    slurm_client = SlurmCLIClient("host")
    return SlurmCLIService(client=slurm_client, store=analysis_store)


@pytest.fixture
def job_service(
    analysis_store: Store,
    slurm_service: SlurmCLIService,
    tower_service: TowerAPIService,
) -> JobService:
    return JobService(
        slurm_service=slurm_service,
        tower_service=tower_service,
        store=analysis_store,
    )


@pytest.fixture
def encryption_service() -> EncryptionService:
    key: bytes = os.urandom(32)
    secret_key: str = base64.b64encode(key).decode()
    return EncryptionService(secret_key)


@pytest.fixture
def google_oauth_client(google_oauth_response: dict, mock_request: Mocker) -> GoogleOAuthClient:
    oauth_base_url = "https://oauth2.googleapis.com/token"
    mock_request.post(oauth_base_url, json=google_oauth_response)

    return GoogleOAuthClient(
        client_id="client_id",
        client_secret="client_secret",
        redirect_uri="redirect_uri",
        oauth_base_url=oauth_base_url,
    )


@pytest.fixture
def google_user_info_response(user_email: str) -> dict:
    return {"email": f"{user_email}"}


@pytest.fixture
def google_api_client(mock_request: Mocker, google_user_info_response: dict) -> GoogleAPIClient:
    base_url = "https://www.googleapis.com"
    user_info_endpoint = f"{base_url}/oauth2/v1/userinfo"
    mock_request.get(user_info_endpoint, json=google_user_info_response)
    return GoogleAPIClient("https://www.googleapis.com")


@pytest.fixture
def authentication_service(
    encryption_service: EncryptionService,
    google_oauth_client: GoogleOAuthClient,
    google_api_client: GoogleAPIClient,
    user_store: Store,
) -> AuthenticationService:
    return AuthenticationService(
        encryption_service=encryption_service,
        store=user_store,
        google_oauth_client=google_oauth_client,
        google_api_client=google_api_client,
    )


@pytest.fixture
def google_oauth_response() -> dict:
    return {
        "access_token": "access_token",
        "token_type": "token_type",
        "expires_in": 3600,
        "refresh_token": "refresh_token",
        "scope": "scope",
        "id_token": "id_token",
    }


@pytest.fixture
def mock_request():
    with Mocker() as mock:
        yield mock



from pathlib import Path
from trailblazer.clients.tower.models import TowerTask
from trailblazer.constants import FileFormat
from trailblazer.io.controller import ReadFile
from trailblazer.store.models import Analysis, Job


def get_tower_workflow_id(analysis: Analysis) -> str:
    file = Path(analysis.config_path)
    content: dict = ReadFile.get_content_from_file(file_format=FileFormat.YAML, file_path=file)
    return content.get(analysis.case_id)[-1]


def create_job_from_tower_task(task: TowerTask) -> Job:
    return Job(
        slurm_id=task.nativeId,
        name=task.process,
        status=task.status,
        started_at=task.start,
        elapsed=int(task.duration / 60),
    )



from pathlib import Path
from trailblazer.constants import FileFormat, JobType, SlurmJobStatus, TrailblazerStatus
from trailblazer.io.controller import ReadFile
from trailblazer.store.models import Analysis, Job


def get_tower_workflow_id(analysis: Analysis) -> str:
    file = Path(analysis.config_path)
    content: dict = ReadFile.get_content_from_file(file_format=FileFormat.YAML, file_path=file)
    return content.get(analysis.case_id)[-1]


def get_status(jobs: list[Job]) -> TrailblazerStatus:
    if has_same_status(jobs):
        return get_single_status(jobs)
    is_running: bool = is_analysis_ongoing(jobs)
    is_failed: bool = is_analysis_failed(jobs)

    if is_failed and is_running:
        return TrailblazerStatus.ERROR

    if is_failed and not is_running:
        return TrailblazerStatus.FAILED

    if is_running:
        return TrailblazerStatus.RUNNING

    return TrailblazerStatus.CANCELLED


def has_same_status(jobs: list[Job]) -> bool:
    if not jobs:
        return False
    first_status = jobs[0].status
    return all(job.status == first_status for job in jobs)


def get_single_status(jobs: list[Job]) -> SlurmJobStatus:
    single_status: str = jobs[0].status
    return (
        TrailblazerStatus.FAILED
        if single_status == SlurmJobStatus.TIME_OUT
        else TrailblazerStatus[single_status.upper()]
    )


def is_analysis_ongoing(jobs: list[Job]) -> bool:
    run_statuses = SlurmJobStatus.ongoing_statuses()
    return any(job.status in run_statuses for job in jobs)


def is_analysis_failed(jobs: list[Job]) -> bool:
    fail_statuses = SlurmJobStatus.fail_statuses()
    return any(job.status in fail_statuses for job in jobs)


def get_progress(jobs: list[Job]) -> float:
    analysis_jobs: list[Job] = _get_analysis_jobs(jobs)
    completed_jobs: list[Job] = _get_completed_jobs(analysis_jobs)

    total_count: int = len(analysis_jobs)
    completed_count: int = len(completed_jobs)

    return 0.0 if total_count == 0 else completed_count / total_count


def _get_analysis_jobs(jobs: list[Job]) -> list[Job]:
    return [job for job in jobs if job.job_type == JobType.ANALYSIS]


def _get_completed_jobs(jobs: list[Job]) -> list[Job]:
    return [job for job in jobs if job.status == SlurmJobStatus.COMPLETED]



from trailblazer.clients.tower.models import TowerTasksResponse
from trailblazer.services.tower.tower_api_service import TowerAPIService
from trailblazer.store.models import Analysis


def test_update_jobs(
    tower_service: TowerAPIService,
    tower_analysis: Analysis,
    tower_tasks_response: TowerTasksResponse,
):
    # GIVEN an analysis started in tower without any jobs

    # GIVEN that some tasks for it have started running
    tower_service.client.get_tasks.return_value = tower_tasks_response

    # WHEN updating the jobs
    tower_service.update_jobs(tower_analysis.id)

    # THEN the jobs should be updated
    assert tower_analysis.jobs



from trailblazer.clients.tower.models import TowerTasksResponse
from trailblazer.clients.tower.tower_client import TowerAPIClient
from trailblazer.constants import TOWER_WORKFLOW_STATUS, TrailblazerStatus
from trailblazer.services.tower.utils import create_job_from_tower_task, get_tower_workflow_id
from trailblazer.store.models import Analysis, Job
from trailblazer.store.store import Store


class TowerAPIService:
    """Class communicating with NF tower regarding a given analysis (workflow)."""

    def __init__(self, client: TowerAPIClient, store: Store) -> None:
        self.client = client
        self.store = store

    def update_jobs(self, analysis_id: int) -> list[Job]:
        analysis: Analysis = self.store.get_analysis_with_id(analysis_id)
        workflow_id: str = get_tower_workflow_id(analysis)
        response: TowerTasksResponse = self.client.get_tasks(workflow_id)
        jobs: list[Job] = [create_job_from_tower_task(task) for task in response.get_tasks()]
        self.store.replace_jobs(analysis_id=analysis_id, jobs=jobs)

    def cancel_jobs(self, analysis_id: int) -> None:
        analysis: Analysis = self.store.get_analysis_with_id(analysis_id)
        workflow_id: str = get_tower_workflow_id(analysis)
        self.client.cancel_workflow(workflow_id)
        self.update_jobs(analysis_id)

    def get_status(self, analysis_id: int) -> TrailblazerStatus:
        analysis: Analysis = self.store.get_analysis_with_id(analysis_id)
        workflow_id: str = get_tower_workflow_id(analysis)
        response = self.client.get_workflow(workflow_id)
        status = TOWER_WORKFLOW_STATUS.get(response.workflow.status, TrailblazerStatus.ERROR)
        if status == TrailblazerStatus.COMPLETED:
            return TrailblazerStatus.QC
        return status



from typing import Generator
from unittest.mock import patch
from flask import Flask
from flask.testing import FlaskClient
import pytest

from trailblazer.server.app import app
from trailblazer.store.store import Store


@pytest.fixture
def flask_app(user_store: Store):
    yield app


@pytest.fixture
def client(flask_app: Flask) -> Generator[FlaskClient, None, None]:
    # Bypass authentication
    with patch.object(flask_app, "before_request_funcs", new={}):
        yield flask_app.test_client()



import json
import pytest
import requests_mock

from trailblazer.clients.slurm_api_client.slurm_api_client import SlurmAPIClient
from trailblazer.clients.tower.tower_client import TowerAPIClient


@pytest.fixture
def slurm_client() -> SlurmAPIClient:
    return SlurmAPIClient(
        base_url="https://slurm.example.com",
        access_token="access_token",
        user_name="user_name",
    )


@pytest.fixture
def tower_client():
    return TowerAPIClient(
        base_url="https://tower/",
        access_token="token",
        workspace_id="workspace_id",
    )


@pytest.fixture
def mock_request():
    with requests_mock.Mocker() as mock:
        yield mock


@pytest.fixture
def jobs_response() -> dict:
    return {"jobs": [{"name": "job 1"}, {"name": "job 2"}]}


@pytest.fixture
def tower_empty_task_response() -> dict:
    return {"tasks": [], "total": 0}


@pytest.fixture
def tower_running_tasks_response() -> dict:
    with open("tests/fixtures/tower/tower_tasks_running.json") as file:
        return json.load(file)


@pytest.fixture
def tower_completed_tasks_response() -> dict:
    with open("tests/fixtures/tower/tower_tasks_completed.json") as file:
        return json.load(file)



import logging
from functools import wraps
from pydantic import ValidationError
from requests import ConnectionError, HTTPError
from requests.exceptions import MissingSchema

from trailblazer.exc import InvalidTowerAPIResponse, TowerAPIClientError, TowerRequestFailed

LOG = logging.getLogger(__name__)


def handle_client_errors(func):
    """Handle errors that may occur when interacting with the Tower API.

    Raises:
        TowerRequestFailed: when the request to the Tower API fails.
        InvalidTowerAPIResponse: when the returned data from the Tower API is invalid.
        TowerAPIClientError: when an unexpected error occurs in the Tower API client.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except (MissingSchema, HTTPError, ConnectionError) as error:
            LOG.error(f"Request against tower failed: {error}, {error.response.text}")
            raise TowerRequestFailed(error) from error
        except ValidationError as error:
            LOG.error(f"Invalidly formatted response from Tower API: {error}")
            raise InvalidTowerAPIResponse(error) from error
        except Exception as error:
            LOG.error(f"Unexpected error in Tower API client: {error}")
            raise TowerAPIClientError(error) from error

    return wrapper



from datetime import datetime
import logging

from trailblazer.constants import TrailblazerStatus, WorkflowManager
from trailblazer.dto import CreateJobRequest, FailedJobsRequest, FailedJobsResponse, JobResponse
from trailblazer.exceptions import JobServiceError, NoJobsError
from trailblazer.services.job_service.mappers import (
    create_failed_jobs_response,
    create_job_response,
)
from trailblazer.services.job_service.utils import (
    get_progress,
    get_status,
)
from trailblazer.services.slurm.dtos import SlurmJobInfo
from trailblazer.services.slurm.slurm_service import SlurmService
from trailblazer.services.tower.tower_api_service import TowerAPIService
from trailblazer.store.models import Analysis, Job
from trailblazer.store.store import Store
from trailblazer.utils.datetime import get_date_number_of_days_ago

LOG = logging.getLogger(__name__)


class JobService:
    def __init__(self, store: Store, slurm_service: SlurmService, tower_service: TowerAPIService):
        self.store = store
        self.slurm_service = slurm_service
        self.tower_service = tower_service

    def get_failed_jobs(self, request: FailedJobsRequest) -> FailedJobsResponse:
        time_window: datetime = get_date_number_of_days_ago(request.days_back)
        failed_jobs: list[dict] = self.store.get_failed_jobs_stats(time_window)
        return create_failed_jobs_response(failed_jobs)

    def add_job(self, analysis_id: int, data: CreateJobRequest) -> JobResponse:
        job: Job = self.store.add_job(analysis_id=analysis_id, job_request=data)
        return create_job_response(job)

    def update_upload_jobs(self) -> None:
        jobs: list[Job] = self.store.get_ongoing_upload_jobs()
        for job in jobs:
            LOG.info(f"Updating upload job {job.id}")
            updated_job: SlurmJobInfo = self.slurm_service.get_job(job.slurm_id)
            self.store.update_job(job_id=job.id, job_info=updated_job)

    def update_jobs(self, analysis_id: int) -> None:
        analysis: Analysis = self.store.get_analysis_with_id(analysis_id)
        try:
            if analysis.workflow_manager == WorkflowManager.SLURM:
                self.slurm_service.update_jobs(analysis_id)
            if analysis.workflow_manager == WorkflowManager.TOWER:
                self.tower_service.update_jobs(analysis_id)
        except Exception as error:
            LOG.error(f"Failed to update jobs {analysis.case_id} - {analysis.id}: {error}")
            raise JobServiceError from error

    def get_analysis_status(self, analysis_id: int) -> TrailblazerStatus:
        analysis: Analysis = self.store.get_analysis_with_id(analysis_id)

        if analysis.status == TrailblazerStatus.CANCELLED:
            return TrailblazerStatus.CANCELLED

        if not analysis.jobs:
            raise NoJobsError(f"No jobs found for analysis {analysis_id}")

        if analysis.workflow_manager == WorkflowManager.TOWER:
            return self.tower_service.get_status(analysis_id)

        return get_status(analysis.jobs)

    def get_analysis_progression(self, analysis_id: int) -> float:
        analysis: Analysis = self.store.get_analysis_with_id(analysis_id)
        return get_progress(analysis.jobs)

    def cancel_jobs(self, analysis_id: int) -> None:
        analysis: int = self.store.get_analysis_with_id(analysis_id)

        if analysis.workflow_manager == WorkflowManager.SLURM:
            self.slurm_service.cancel_jobs(analysis_id)
        if analysis.workflow_manager == WorkflowManager.TOWER:
            self.tower_service.cancel_jobs(analysis_id)



"""Utility functions to simply add test data in a Trailblazer store."""

from datetime import date, datetime
from sqlalchemy.orm import Session

from trailblazer.store.store import Store
from trailblazer.store.database import get_session
from trailblazer.store.models import Info, Job, User


class StoreHelpers:
    """Class to hold helper functions that needs to be used all over."""

    @staticmethod
    def add_info(store: Store) -> Info:
        """Add an info object to the store."""
        session: Session = get_session()
        info: Info = Info()
        session.add(info)
        return info

    @staticmethod
    def add_job(
        analysis_id: int,
        name: str,
        slurm_id: int,
        status: str,
        started_at: date = datetime.now(),
    ) -> Job:
        """Add an job object to the store."""
        session: Session = get_session()
        job: Job = Job(
            analysis_id=analysis_id,
            slurm_id=slurm_id,
            name=name,
            started_at=started_at,
            status=status,
        )
        session.add(job)
        return job

    @staticmethod
    def add_user(email: str, name: str, abbreviation: str, is_archived: bool = False) -> User:
        """Add a user object to the store."""
        session: Session = get_session()
        user = User(email=email, name=name, abbreviation=abbreviation, is_archived=is_archived)
        session.add(user)
        return user



from trailblazer.clients.slurm_api_client.dto.job_response import SlurmJobResponse
from trailblazer.clients.slurm_api_client.slurm_api_client import SlurmAPIClient
from trailblazer.services.slurm.dtos import SlurmJobInfo
from trailblazer.services.slurm.slurm_api_service.mappers import create_job
from trailblazer.services.slurm.slurm_api_service.utils import create_job_info_dto
from trailblazer.services.slurm.slurm_service import SlurmService
from trailblazer.services.slurm.utils import get_slurm_job_ids
from trailblazer.store.models import Analysis, Job
from trailblazer.store.store import Store


class SlurmAPIService(SlurmService):
    def __init__(self, client: SlurmAPIClient, store: Store):
        self.client = client
        self.store = store

    def get_job(self, job_id: int) -> SlurmJobInfo:
        job_response: SlurmJobResponse = self.client.get_job(job_id)
        return create_job_info_dto(job_response)

    def update_jobs(self, analysis_id: int) -> None:
        analysis: Analysis = self.store.get_analysis_with_id(analysis_id)
        job_ids: list[int] = get_slurm_job_ids(analysis.config_path)

        dtos: list[SlurmJobInfo] = []
        for job_id in job_ids:
            if job := self.get_job(job_id):
                dtos.append(job)

        jobs: list[Job] = [create_job(dto) for dto in dtos]
        self.store.replace_jobs(analysis_id=analysis_id, jobs=jobs)

    def cancel_jobs(self, analysis_id: int) -> None:
        analysis: Analysis = self.store.get_analysis_with_id(analysis_id)
        jobs: list[Job] = analysis.jobs

        for job in jobs:
            self.client.cancel_job(job.slurm_id)
        self.update_jobs(analysis_id)



from requests_mock import Mocker
from trailblazer.clients.tower.tower_client import TowerAPIClient


def test_empty_tasks_response_is_accepted(
    tower_client: TowerAPIClient,
    tower_empty_task_response: dict,
    mock_request: Mocker,
):
    # GIVEN a workflow response without tasks
    mock_request.get("https://tower/workflow/1/tasks", json=tower_empty_task_response)

    # WHEN fetching the tasks for a workflow
    response = tower_client.get_tasks(workflow_id="1")

    # THEN the response should not contain any tasks
    assert response.tasks == []

    # THEN the response should have a total of 0 tasks
    assert response.total == 0


def test_running_tasks_response_is_accepted(
    tower_client: TowerAPIClient,
    tower_running_tasks_response: dict,
    mock_request: Mocker,
):
    # GIVEN a workflow response with running tasks
    mock_request.get("https://tower/workflow/1/tasks", json=tower_running_tasks_response)

    # WHEN fetching the tasks for a workflow
    response = tower_client.get_tasks(workflow_id="1")

    # THEN the response should contain running tasks
    assert response.tasks

    # THEN the response should have a total of tasks
    assert response.total


def test_completed_tasks_response_is_accepted(
    tower_client: TowerAPIClient,
    tower_completed_tasks_response: dict,
    mock_request: Mocker,
):
    # GIVEN a workflow response with completed tasks
    mock_request.get("https://tower/workflow/1/tasks", json=tower_completed_tasks_response)

    # WHEN fetching the tasks for a workflow
    response = tower_client.get_tasks(workflow_id="1")

    # THEN the response should contain completed tasks
    assert response.tasks

    # THEN the response should have a total of tasks
    assert response.total



from trailblazer.apps.slurm.models import SqueueJob, SqueueResult
from trailblazer.clients.slurm_cli_client.slurm_cli_client import SlurmCLIClient
from trailblazer.clients.slurm_cli_client.utils import create_job_info_dto
from trailblazer.services.slurm.dtos import SlurmJobInfo
from trailblazer.services.slurm.slurm_api_service.mappers import create_job
from trailblazer.services.slurm.slurm_service import SlurmService
from trailblazer.services.slurm.utils import get_slurm_job_ids
from trailblazer.store.models import Analysis
from trailblazer.store.store import Store


class SlurmCLIService(SlurmService):
    def __init__(self, client: SlurmCLIClient, store: Store):
        self.client = client
        self.store = store

    def get_job(self, job_id: int) -> SlurmJobInfo:
        queue: SqueueResult = self.client.get_slurm_queue([job_id])
        job: SqueueJob = queue.jobs[0]
        return create_job_info_dto(job)

    def update_jobs(self, analysis_id: int) -> None:
        analysis: Analysis = self.store.get_analysis_with_id(analysis_id)
        job_ids: list[int] = get_slurm_job_ids(analysis.config_path)
        queue: SqueueResult = self.client.get_slurm_queue(job_ids)
        dtos = [create_job_info_dto(job) for job in queue.jobs]
        jobs = [create_job(dto) for dto in dtos]
        self.store.replace_jobs(analysis_id=analysis_id, jobs=jobs)

    def cancel_jobs(self, analysis_id: int) -> None:
        analysis: Analysis = self.store.get_analysis_with_id(analysis_id)
        job_ids: list[int] = get_slurm_job_ids(analysis.config_path)
        for job_id in job_ids:
            self.client.cancel_job(job_id)
        self.update_jobs(analysis_id)



from pathlib import Path
from typing import Any, Callable

from trailblazer.constants import FileFormat
from trailblazer.io.csv import read_csv_stream
from trailblazer.io.json import read_json
from trailblazer.io.yaml import read_yaml


class ReadFile:
    """Reading file using different methods."""

    read_file = {
        FileFormat.JSON: read_json,
        FileFormat.YAML: read_yaml,
    }

    @classmethod
    def get_content_from_file(cls, file_format: str, file_path: Path) -> Any:
        """Read file using file format dispatch table."""
        return cls.read_file[file_format](file_path=file_path)


class ReadStream:
    """Reading stream using different methods."""

    read_stream: dict[str, Callable] = {
        FileFormat.CSV: read_csv_stream,
    }

    @classmethod
    def get_content_from_stream(cls, file_format: str, stream: Any, **kwargs) -> Any:
        """Read stream using file format dispatch table."""
        return cls.read_stream[file_format](stream=stream, **kwargs)



import datetime
from http import HTTPStatus
from flask import Request, jsonify
import logging
from functools import wraps
from pydantic import ValidationError

from trailblazer.exc import CancelSlurmAnalysisNotSupportedError, MissingAnalysis
from trailblazer.dto import AnalysesRequest
from trailblazer.services.authentication_service.exceptions import AuthenticationError

LOG = logging.getLogger(__name__)


def parse_analyses_request(request: Request) -> AnalysesRequest:
    """Parse a request for retrieving analyses."""
    query_params = {}
    for key in request.args.keys():
        if key_has_list_of_values(key):
            query_params[key[:-2]] = request.args.getlist(key)
        else:
            query_params[key] = request.args.get(key)
    return AnalysesRequest.model_validate(query_params)


def key_has_list_of_values(key: str) -> bool:
    return key.endswith("[]")


def stringify_timestamps(data: dict) -> dict[str, str]:
    """Convert datetime into string before dumping in order to avoid information loss"""
    for key, val in data.items():
        if isinstance(val, datetime.datetime):
            data[key] = str(val)
    return data


def handle_endpoint_errors(func):

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except AuthenticationError:
            return jsonify("User not allowed"), HTTPStatus.FORBIDDEN
        except MissingAnalysis as error:
            return jsonify(error=str(error)), HTTPStatus.NOT_FOUND
        except CancelSlurmAnalysisNotSupportedError as error:
            return jsonify(error=str(error)), HTTPStatus.BAD_REQUEST
        except ValidationError as error:
            LOG.error(f"Validation error in analysis endpoint: {error}")
            return jsonify(error=str(error)), HTTPStatus.BAD_REQUEST
        except Exception as error:
            LOG.error(f"Unexpected error in analysis endpoint: {error}")
            return (
                jsonify(error="An error occurred while processing your request."),
                HTTPStatus.INTERNAL_SERVER_ERROR,
            )

    return wrapper



from trailblazer.constants import TrailblazerStatus
from trailblazer.dto.update_analyses import AnalysisUpdate, UpdateAnalyses
from trailblazer.services.analysis_service.analysis_service import AnalysisService
from trailblazer.store.models import Analysis, User
from trailblazer.store.store import Store


def test_patch_analyses_delivered(
    analysis_store: Store,
    analysis_id: int,
    analysis_service: AnalysisService,
):
    # GIVEN a request to mark an analysis as delivered
    analysis: Analysis = analysis_store.get_analysis_with_id(analysis_id)
    analysis_update = AnalysisUpdate(id=analysis_id, is_delivered=True)
    update_analyses = UpdateAnalyses(analyses=[analysis_update])
    user = User(id=1)

    # WHEN updating the analysis
    analysis_service.update_analyses(data=update_analyses, user=user)

    # THEN the analysis should be marked as delivered
    assert analysis.delivered_by


def test_cancel_analysis(analysis_service: AnalysisService, analysis_with_running_jobs: Analysis):
    # GIVEN a running analysis

    # WHEN cancelling the analysis
    analysis_service.cancel_analysis(analysis_with_running_jobs.id)

    # THEN the analysis should be cancelled
    assert analysis_with_running_jobs.status == TrailblazerStatus.CANCELLED

    # THEN a comment should be added to the analysis
    assert analysis_with_running_jobs.comment


def test_update_analysis_meta_data(
    analysis_service: AnalysisService,
    analysis_with_running_jobs: Analysis,
):
    # GIVEN that the associated jobs are completed
    analysis_service.job_service.get_analysis_status.return_value = TrailblazerStatus.COMPLETED
    analysis_service.job_service.get_analysis_progression.return_value = 100

    # WHEN updating the analysis
    analysis_service.update_analysis_meta_data(analysis_with_running_jobs.id)

    # THEN the analysis should be completed
    assert analysis_with_running_jobs.status == TrailblazerStatus.COMPLETED
    assert analysis_with_running_jobs.progress == 100



from pydantic import BaseModel


class StatusSummary(BaseModel):
    count: int = 0
    case_ids: list[str] = []


class Summary(BaseModel):
    order_id: int
    total: int

    cancelled: StatusSummary
    completed: StatusSummary
    delivered: StatusSummary
    failed: StatusSummary
    running: StatusSummary


class SummariesResponse(BaseModel):
    summaries: list[Summary]



import logging
from pydantic import ValidationError
import requests

from trailblazer.clients.slurm_api_client.dto import SlurmJobResponse
from trailblazer.exc import ResponseDeserializationError, SlurmAPIClientError


LOG = logging.getLogger(__name__)


class SlurmAPIClient:
    def __init__(self, base_url: str, access_token: str, user_name: str) -> None:
        self.base_url = base_url
        self.headers = {"X-SLURM-USER-NAME": user_name, "X-SLURM-USER-TOKEN": access_token}

    def get_job(self, job_id: str) -> SlurmJobResponse:
        endpoint: str = f"{self.base_url}/slurm/v0.0.40/job/{job_id}"
        try:
            response = requests.get(endpoint, headers=self.headers)
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            LOG.error(f"Error getting job {job_id}: {e.response.content}")
            raise SlurmAPIClientError(e)

        try:
            return SlurmJobResponse.model_validate(response.json())
        except ValidationError as e:
            LOG.error(f"Error deserializing job response: {e}")
            raise ResponseDeserializationError(e)

    def cancel_job(self, job_id: str) -> None:
        endpoint: str = f"{self.base_url}/slurm/v0.0.40/job/{job_id}"
        try:
            response = requests.delete(endpoint, headers=self.headers)
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            LOG.error(f"Error cancelling job {job_id}: {e.response.content}")
            raise SlurmAPIClientError(e)



from datetime import datetime, timedelta
from trailblazer.constants import SlurmJobStatus, TrailblazerStatus, Workflow
from trailblazer.dto.analyses_response import UpdateAnalysesResponse
from trailblazer.dto.analysis_response import AnalysisResponse
from trailblazer.dto.summaries_response import StatusSummary, Summary
from trailblazer.store.models import Analysis
from trailblazer.store.models import Job


def get_status_counts(analyses: list[Analysis]) -> dict[TrailblazerStatus, StatusSummary]:
    """Returns the number of analyses with each status."""
    delivered: int = 0
    delivered_cases: list[str] = []
    status_counts: dict = {status: StatusSummary() for status in TrailblazerStatus}
    for analysis in analyses:
        if analysis.delivery:
            delivered += 1
            delivered_cases.append(analysis.case_id)
        else:
            status_counts[analysis.status].count += 1
            status_counts[analysis.status].case_ids.append(analysis.case_id)
    status_counts["delivered"] = StatusSummary(count=delivered, case_ids=delivered_cases)
    return status_counts


def create_summary(analyses: list[Analysis], order_id: int) -> Summary:
    total: int = len(analyses)
    status_counts: dict[TrailblazerStatus, StatusSummary] = get_status_counts(analyses)
    return Summary(order_id=order_id, total=total, **status_counts)


def create_analysis_response(analysis: Analysis) -> AnalysisResponse:
    analysis_data: dict = analysis.to_dict()
    analysis_data["jobs"] = [job.to_dict() for job in analysis.analysis_jobs]
    analysis_data["upload_jobs"] = [job.to_dict() for job in analysis.upload_jobs]
    analysis_data["user"] = analysis.user.to_dict() if analysis.user else None
    return AnalysisResponse.model_validate(analysis_data)


def create_update_analyses_response(analyses: list[Analysis]) -> UpdateAnalysesResponse:
    response_data: list[dict] = []
    for analysis in analyses:
        analysis_data = analysis.to_dict()
        response_data.append(analysis_data)
    return UpdateAnalysesResponse(analyses=response_data)


def get_upload_date(analysis: Analysis) -> datetime | None:
    completed_job: Job | None = _get_completed_job(analysis.upload_jobs)
    if not completed_job:
        return
    return _get_completed_at_date(completed_job)


def _get_completed_job(jobs: list[Job]) -> Job | None:
    for job in jobs:
        if _is_completed_job(job):
            return job


def _is_completed_job(job: Job):
    return job.status == SlurmJobStatus.COMPLETED and job.started_at and job.elapsed is not None


def _get_completed_at_date(job: Job) -> datetime:
    return job.started_at + timedelta(seconds=job.elapsed)



import datetime
import uuid

from sqlalchemy import Column, ForeignKey, UniqueConstraint, orm, types

from trailblazer.constants import (
    PRIORITY_OPTIONS,
    TYPES,
    JobType,
    SlurmJobStatus,
    TrailblazerStatus,
    WorkflowManager,
)
from trailblazer.store.database import Model


class Info(Model):
    """Keep track of metadata."""

    __tablename__ = "info"

    id = Column(types.Integer, primary_key=True)
    created_at = Column(types.DateTime, default=datetime.datetime.now)
    updated_at = Column(types.DateTime)

    def to_dict(self) -> dict:
        """Return a dictionary representation of the object."""
        return {
            "id": self.id,
            "created_at": self.created_at,
            "updated_at": self.updated_at,
        }


class User(Model):
    __tablename__ = "user"

    created_at = Column(types.DateTime, default=datetime.datetime.now)
    email = Column(types.String(128), unique=True)
    google_id = Column(types.String(128), unique=True)
    id = Column(types.Integer, primary_key=True)
    is_archived = Column(types.Boolean, default=False)
    name = Column(types.String(128))
    abbreviation = Column(types.String(32), unique=True, nullable=False)
    refresh_token = Column(types.Text)

    runs = orm.relationship("Analysis", backref="user")

    @property
    def first_name(self) -> str:
        """First part of name."""
        return self.name.split(" ")[0]

    def __str__(self) -> str:
        return self.name

    def to_dict(self) -> dict:
        return {
            "id": self.id,
            "name": self.name,
            "email": self.email,
            "created_at": self.created_at,
            "is_archived": self.is_archived,
        }


class Analysis(Model):
    """Analysis record."""

    __tablename__ = "analysis"
    __table_args__ = (
        UniqueConstraint("case_id", "started_at", "status", name="_uc_case_id_start_status"),
    )

    case_id = Column(types.String(128), nullable=False)
    comment = Column(types.Text)
    completed_at = Column(types.DateTime)
    config_path = Column(types.Text)
    id = Column(types.Integer, primary_key=True)
    is_visible = Column(types.Boolean, default=True)
    logged_at = Column(types.DateTime, default=datetime.datetime.now)
    order_id = Column(types.Integer, index=True)
    out_dir = Column(types.Text)
    priority = Column(types.Enum(*PRIORITY_OPTIONS))
    progress = Column(types.Float, default=0.0)
    started_at = Column(types.DateTime)
    status = Column(types.Enum(*TrailblazerStatus.statuses()))
    ticket_id = Column(types.String(32))
    type = Column(types.Enum(*TYPES))
    uploaded_at = Column(types.DateTime)
    user_id = Column(ForeignKey(User.id))
    version = Column(types.String(32))
    workflow = Column(types.String(32))
    workflow_manager = Column(types.Enum(*WorkflowManager.list()), default=WorkflowManager.SLURM)

    jobs = orm.relationship("Job", cascade="all,delete", backref="analysis")
    delivery = orm.relationship("Delivery", uselist=False)

    @property
    def has_ongoing_status(self) -> bool:
        """Check if analysis status is ongoing."""
        return self.status in TrailblazerStatus.ongoing_statuses()

    @property
    def analysis_jobs(self) -> list["Job"]:
        """Return upload jobs."""
        return [job for job in self.jobs if job.job_type == JobType.ANALYSIS]

    @property
    def upload_jobs(self) -> list["Job"]:
        """Return upload jobs."""
        return [job for job in self.jobs if job.job_type == JobType.UPLOAD]

    @property
    def delivered_by(self) -> str | None:
        return self.delivery.user.abbreviation if self.delivery else None

    @property
    def delivered_date(self) -> datetime.datetime | None:
        return self.delivery.delivered_date if self.delivery else None

    def to_dict(self) -> dict:
        """Return a dictionary representation of the object."""
        return {
            "id": self.id,
            "case_id": self.case_id,
            "is_delivered": bool(self.delivery),
            "delivered_by": self.delivered_by,
            "delivered_date": self.delivered_date,
            "version": self.version,
            "logged_at": self.logged_at,
            "started_at": self.started_at,
            "completed_at": self.completed_at,
            "status": self.status,
            "priority": self.priority,
            "order_id": self.order_id,
            "out_dir": self.out_dir,
            "config_path": self.config_path,
            "comment": self.comment,
            "is_visible": self.is_visible,
            "type": self.type,
            "user_id": self.user_id,
            "progress": self.progress,
            "workflow": self.workflow,
            "ticket_id": self.ticket_id,
            "uploaded_at": self.uploaded_at,
            "workflow_manager": self.workflow_manager,
        }


class Delivery(Model):
    """Tracks when and by whom an analysis was delivered."""

    __tablename__ = "delivery"
    __table_args__ = (UniqueConstraint("analysis_id"),)

    id = Column(types.Uuid, primary_key=True, default=uuid.uuid4)
    analysis_id = Column(ForeignKey(Analysis.id, ondelete="CASCADE"), nullable=False)
    delivered_by = Column(ForeignKey(User.id), nullable=False)
    delivered_date = Column(types.Date, nullable=False)

    analysis = orm.relationship("Analysis", foreign_keys=[analysis_id], back_populates="delivery")
    user = orm.relationship("User", foreign_keys=[delivered_by])


class Job(Model):
    """Represent a step in the workflow."""

    __tablename__ = "job"

    id = Column(types.Integer, primary_key=True)
    analysis_id = Column(ForeignKey(Analysis.id, ondelete="CASCADE"), nullable=False)
    slurm_id = Column(types.Integer)
    name = Column(types.String(64))
    context = Column(types.String(64))
    started_at = Column(types.DateTime)
    elapsed = Column(types.Integer)
    status = Column(types.Enum(*SlurmJobStatus.statuses()))
    job_type = Column(types.Enum(*JobType.types()), default=JobType.ANALYSIS, nullable=False)

    def to_dict(self) -> dict:
        """Return a dictionary representation of the object."""
        return {
            "id": self.id,
            "analysis_id": self.analysis_id,
            "slurm_id": self.slurm_id,
            "name": self.name,
            "context": self.context,
            "started_at": self.started_at,
            "elapsed": self.elapsed,
            "status": self.status,
        }



from datetime import datetime

import pytest

from tests.mocks.store_mock import MockStore
from tests.store.utils.store_helper import StoreHelpers
from trailblazer.constants import TrailblazerStatus
from trailblazer.exc import MissingAnalysis
from trailblazer.store.models import Analysis, Job, User


def test_get_analyses_by_status_started_at_and_comment(
    analysis_store: MockStore, timestamp_now: datetime
):
    """Test getting an analysis when it fulfills all criteria."""
    # GIVEN a store with an analysis
    existing_analysis: Analysis = analysis_store.get_query(table=Analysis).first()

    # GIVEN a comment
    analysis_store.update_latest_analysis_comment(
        case_id=existing_analysis.case_id, comment="a new comment"
    )

    # WHEN getting analysis that fulfills criteria
    analyses: list[Analysis] = analysis_store.get_analyses_by_status_started_at_and_comment(
        status=existing_analysis.status,
        before=timestamp_now,
        comment=existing_analysis.comment,
    )

    # THEN it should return the same analysis
    assert analyses[0] == existing_analysis


def test_get_analyses_by_status_started_at_and_comment_with_comment(analysis_store: MockStore):
    """Test getting an analysis when only supplying comment."""
    # GIVEN a store with an analysis
    existing_analysis: Analysis = analysis_store.get_query(table=Analysis).first()

    # GIVEN a comment
    analysis_store.update_latest_analysis_comment(
        case_id=existing_analysis.case_id, comment="a new comment"
    )

    # WHEN getting analysis that fulfills criteria
    analyses: list[Analysis] = analysis_store.get_analyses_by_status_started_at_and_comment(
        comment=existing_analysis.comment,
    )

    # THEN it should return the same analysis
    assert analyses[0] == existing_analysis


def test_get_analyses_by_status_started_at_and_comment_with_status(analysis_store: MockStore):
    """Test getting an analysis when only supplying status."""
    # GIVEN a store with an analysis
    existing_analysis: Analysis = analysis_store.get_query(table=Analysis).first()

    # WHEN getting analysis that fulfills criteria
    analyses: list[Analysis] = analysis_store.get_analyses_by_status_started_at_and_comment(
        status=existing_analysis.status,
    )

    # THEN it should return analyses with supplied status
    for analysis in analyses:
        assert analysis.status == TrailblazerStatus.PENDING


def test_get_analyses_by_status_started_at_and_comment_with_before(
    analysis_store: MockStore, timestamp_now: datetime
):
    """Test getting an analysis when only supplying before."""
    # GIVEN a store with an analysis

    # WHEN getting analysis that fulfills criteria
    analyses: list[Analysis] = analysis_store.get_analyses_by_status_started_at_and_comment(
        before=timestamp_now,
    )

    # THEN it should return analyses started before the supplied date
    for analysis in analyses:
        assert analysis.started_at < timestamp_now


def test_get_analysis(analysis_store: MockStore):
    """Test getting an analysis when it exists in the database."""
    # GIVEN a store with an analysis
    existing_analysis: Analysis = analysis_store.get_query(table=Analysis).first()

    # WHEN getting analysis
    analysis: Analysis = analysis_store.get_analysis(
        case_id=existing_analysis.case_id,
        started_at=existing_analysis.started_at,
        status=existing_analysis.status,
    )

    # THEN it should return the same analysis
    assert analysis == existing_analysis


def test_get_latest_analysis_for_case(analysis_store: MockStore):
    """Test getting an analysis fora case when it exists in the database."""
    # GIVEN a store with an analysis
    existing_analysis: Analysis = analysis_store.get_query(table=Analysis).first()

    # WHEN accessing it by case id
    analysis: Analysis | None = analysis_store.get_latest_analysis_for_case(
        case_id=existing_analysis.case_id
    )

    # THEN it should return the same analysis
    assert analysis == existing_analysis


def test_get_latest_analysis_for_case_when_missing(
    analysis_store: MockStore, case_id_not_in_db: str
):
    """Test getting an analysis fora case when it does not exist in the database."""

    # WHEN accessing it by case id
    analysis: Analysis | None = analysis_store.get_latest_analysis_for_case(
        case_id=case_id_not_in_db
    )

    # THEN no analysis should be returned
    assert not analysis


def test_get_analysis_with_id(analysis_store: MockStore):
    """Test getting an analysis by id when it exists in the database."""
    # GIVEN a store with an analysis
    existing_analysis: Analysis = analysis_store.get_query(table=Analysis).first()

    # WHEN accessing it by ID
    analysis: Analysis | None = analysis_store.get_analysis_with_id(
        analysis_id=existing_analysis.id
    )

    # THEN it should return the same analysis
    assert analysis == existing_analysis


def test_get_analysis_with_id_when_missing(analysis_store: MockStore):
    """Test getting an analysis by database entry id when it does not exist in the database."""
    # GIVEN an id that doesn't exist
    missing_analysis_id: int = 12312423534

    # WHEN accessing the analysis
    with pytest.raises(MissingAnalysis):
        analysis_store.get_analysis_with_id(analysis_id=missing_analysis_id)


def test_get_analyses_for_case(analysis_store: MockStore, case_id: str):
    """Test getting analyses for a case when it exists in the database."""
    # GIVEN a store with an analysis

    # WHEN accessing it by case id
    analyses: list[Analysis] | None = analysis_store.get_analyses_for_case(case_id=case_id)

    # THEN it should return the analyses
    assert analyses


def test_get_analyses_with_statuses(analysis_store: MockStore):
    """Test getting analyses with status when it exists in the database."""
    # GIVEN a store with an analysis

    # WHEN accessing an analysis by statuses
    analyses: list[Analysis] | None = analysis_store.get_analyses_with_statuses(
        statuses=list(TrailblazerStatus.ongoing_statuses())
    )

    # THEN an analysis should be returned
    assert analyses

    # THEN the status should be one of the ongoing ones
    assert analyses[0].status == TrailblazerStatus.PENDING


def test_get_nr_jobs_with_status_per_category(job_store: MockStore, timestamp_yesterday: datetime):
    """Test getting the number of failed jobs per category since a supplied date from the database."""
    # GIVEN a database with jobs

    # WHEN querying for failed users
    failed_jobs: list[dict[str, str | int]] = job_store.get_nr_jobs_with_status_per_category(
        since_when=timestamp_yesterday, status=TrailblazerStatus.FAILED
    )

    # THEN failed jobs should be returned
    assert failed_jobs

    # THEN one job should be returned
    assert len(failed_jobs) == 1

    # THEN name and count should be defined
    assert failed_jobs[0].get("name") == "1"
    assert failed_jobs[0].get("count") == 1


def test_get_user(user_store: MockStore, user_email: str):
    """Test getting a user from the database."""
    # GIVEN a database with a user

    # WHEN querying for a user
    user: User = user_store.get_user(email=user_email)

    # THEN it should be returned
    assert user.email == user_email


def test_get_user_including_archived(user_store: MockStore, archived_user_email: str):
    """Test getting an archived user from the database."""
    # GIVEN a database with an archived user

    # WHEN querying for a user
    user: User = user_store.get_user(email=archived_user_email, exclude_archived=False)

    # THEN it should be returned
    assert user.email == archived_user_email


def test_get_user_including_archive_false(user_store: MockStore, archived_user_email: str):
    """Test getting an archived user from the database."""
    # GIVEN a database with an archived user

    # WHEN querying for a user
    user: User = user_store.get_user(email=archived_user_email, exclude_archived=True)

    # THEN no user should be returned
    assert not user


def test_get_user_when_non_existing(user_store: MockStore):
    """Test getting a non-existing user from the database."""
    # GIVEN a database with a user

    # WHEN querying for a user that doesn't exist
    user: User = user_store.get_user(email="this_is_a_made_up_email@fake_example.com")

    # THEN it should return as None
    assert user is None


def test_get_users(user_store: MockStore, user_email: str, username: str):
    """Test getting a user with username and email."""
    # GIVEN a database with a user

    # WHEN getting users
    users: list[User] = user_store.get_users(name=username, email=user_email)

    # THEN the user should be returned
    assert users[0].email == user_email


def test_get_users_including_archived(user_store: MockStore, user_email: str, username: str):
    """Test getting a user with username and email which are archived."""
    # GIVEN a database with a user and an archived user
    StoreHelpers.add_user(
        name=username,
        email="old.user@magnolia.com",
        is_archived=True,
        abbreviation="NU",
    )

    # WHEN getting users
    users: list[User] = user_store.get_users(name=username, exclude_archived=False)

    # THEN the two users should be returned
    assert len(users) == 2


def test_get_users_no_username(user_store: MockStore, user_email: str):
    """Test getting a user by email."""
    # GIVEN a database with a user

    # WHEN getting users
    users: list[User] = user_store.get_users(email=user_email)

    # THEN the user should be returned
    assert users[0].email == user_email


def test_get_latest_failed_job_for_analysis(job_store: MockStore):
    """Test getting the latest failed job for an analysis."""
    # GIVEN a database with a job and an analysis
    analysis_id = 1

    # WHEN getting the latest failed job
    job: Job | None = job_store.get_latest_failed_job_for_analysis(analysis_id)

    # THEN the job should be returned
    assert job.analysis_id == analysis_id
    assert job.status == TrailblazerStatus.FAILED


def test_get_latest_failed_job_when_none_failed(job_store: MockStore):
    # GIVEN a database with an analysis without failed jobs
    analysis_id = 0

    # WHEN getting the latest failed job
    job: Job | None = job_store.get_latest_failed_job_for_analysis(analysis_id)

    # THEN no job should be returned
    assert not job



from datetime import datetime

from sqlalchemy.orm import Query

from tests.mocks.store_mock import MockStore
from trailblazer.constants import TrailblazerStatus
from trailblazer.store.filters.job_filters import (
    filter_jobs_by_analysis_id,
    filter_jobs_by_status,
    filter_jobs_by_since_when,
)
from trailblazer.store.models import Job


def test_filter_jobs_by_since_when(job_store: MockStore, timestamp_yesterday: datetime):
    """Test return jobs started after the given date."""
    # GIVEN a store containing jobs

    # WHEN retrieving a job by date
    jobs: Query = filter_jobs_by_since_when(
        jobs=job_store.get_query(table=Job), since_when=timestamp_yesterday
    )

    # ASSERT that the jobs is a query
    assert isinstance(jobs, Query)

    # THEN the jobs should have benn started after the given date
    for job in jobs:
        assert job.started_at > timestamp_yesterday


def test_filter_jobs_by_since_when_with_current_date(job_store: MockStore, timestamp_now: datetime):
    """Test return jobs started before the given date."""
    # GIVEN a store containing jobs

    # WHEN retrieving a job by date
    jobs: Query = filter_jobs_by_since_when(
        jobs=job_store.get_query(table=Job), since_when=timestamp_now
    )

    # ASSERT that no jobs are returned
    assert not jobs.count()


def test_filter_jobs_by_status(job_store: MockStore):
    """Test return jobs by status."""
    # GIVEN a store containing jobs

    # WHEN retrieving a job by status
    jobs: Query = filter_jobs_by_status(
        jobs=job_store.get_query(table=Job), status=TrailblazerStatus.COMPLETED
    )

    # ASSERT that the jobs is a query
    assert isinstance(jobs, Query)

    # THEN the jobs attribute status should match the original
    for job in jobs:
        assert job.status == TrailblazerStatus.COMPLETED


def test_filter_jobs_by_analysis_id(job_store: MockStore):
    """Test return jobs by analysis id."""
    # GIVEN a store containing jobs and an analysis id
    analysis_id = 0

    # WHEN retrieving a job by analysis id
    jobs: Query = filter_jobs_by_analysis_id(
        jobs=job_store.get_query(table=Job), analysis_id=analysis_id
    )

    # ASSERT that the jobs is a query
    assert isinstance(jobs, Query)

    # THEN the jobs attribute analysis id should match the original
    assert jobs[0].analysis_id == analysis_id



from datetime import datetime
from tests.mocks.store_mock import MockStore
from trailblazer.constants import TrailblazerStatus
from trailblazer.store.filters.user_filters import UserFilter, apply_user_filter
from trailblazer.store.models import Analysis, User


def test_update_analysis_jobs(analysis_store: MockStore, tower_jobs: list[dict], case_id: str):
    """Test jobs are successfully updated."""

    # GIVEN an analysis with no jobs
    analysis: Analysis = analysis_store.get_latest_analysis_for_case(case_id=case_id)
    assert not analysis.jobs

    # WHEN jobs are updated
    analysis_store.update_jobs(jobs=tower_jobs[:2])

    # THEN there should be jobs
    assert analysis.jobs


def test_update_user_is_archived(user_store: MockStore, user_email: str):
    """Test updating user is archived attribute."""
    # GIVE a database and a not archived user
    user: User = apply_user_filter(
        filter_functions=[UserFilter.BY_EMAIL],
        users=user_store.get_query(table=User),
        email=user_email,
    ).first()
    assert not user.is_archived

    # WHEN updating a user archive status
    user_store.update_user_is_archived(user=user, archive=True)

    archived_user: User = apply_user_filter(
        filter_functions=[UserFilter.BY_EMAIL],
        users=user_store.get_query(table=User),
        email=user_email,
    ).first()

    # THEN user should be recorded as archived in the database
    assert archived_user


def test_update_analysis_status_with_failed(analysis_store: MockStore, case_id: str):
    """Test setting analysis to failed for an analysis."""

    # GIVEN a store with an analysis
    analysis: Analysis | None = analysis_store.get_latest_analysis_for_case(case_id=case_id)
    assert analysis.status != TrailblazerStatus.FAILED

    # WHEN setting analysis to failed
    analysis_store.update_analysis_status_by_case_id(
        case_id=analysis.case_id, status=TrailblazerStatus.FAILED
    )

    # THEN the analysis status should be updated to failed
    assert analysis.status == TrailblazerStatus.FAILED


def test_update_analysis_status_to_completed(analysis_store: MockStore, case_id: str):
    """Test setting analysis to completed for an analysis."""

    # GIVEN a store with an analysis
    analysis: Analysis | None = analysis_store.get_latest_analysis_for_case(case_id=case_id)
    assert analysis.status != TrailblazerStatus.COMPLETED

    # WHEN setting analysis to completed
    analysis_store.update_analysis_status_to_completed(analysis_id=analysis.id)

    # THEN the analysis status should be updated to completed
    assert analysis.status == TrailblazerStatus.COMPLETED


def test_update_analysis_uploaded_at(
    analysis_store: MockStore, timestamp_now: datetime, case_id: str
):
    """Test setting analysis uploaded at for an analysis."""
    # GIVEN a store with an analysis
    analysis: Analysis | None = analysis_store.get_latest_analysis_for_case(case_id=case_id)

    # WHEN setting an analysis uploaded at
    analysis_store.update_analysis_uploaded_at(case_id=analysis.case_id, uploaded_at=timestamp_now)

    # THEN uploaded at should be updated
    assert analysis.uploaded_at == timestamp_now


def test_update_analysis_comment(analysis_store: MockStore, case_id: str):
    """Test adding comment to an analysis."""

    # GIVEN a store with an analysis
    analysis: Analysis | None = analysis_store.get_latest_analysis_for_case(case_id=case_id)
    comment: str = "test comment"

    # WHEN adding a comment
    analysis_store.update_latest_analysis_comment(case_id=analysis.case_id, comment=comment)

    # THEN a comment should have been added
    assert analysis.comment == comment


def test_update_analysis_comment_when_existing(analysis_store: MockStore, case_id: str):
    """Test adding comment to an analysis when a comment already exists."""

    # GIVEN a store with an analysis
    analysis: Analysis | None = analysis_store.get_latest_analysis_for_case(case_id=case_id)
    first_comment: str = "One"
    second_comment: str = "Second"

    # WHEN adding a comment
    analysis_store.update_latest_analysis_comment(case_id=analysis.case_id, comment=first_comment)
    analysis_store.update_latest_analysis_comment(case_id=analysis.case_id, comment=second_comment)

    # THEN comments should have been added
    assert analysis.comment == f"{first_comment} {second_comment}"


def test_update_tower_jobs(analysis_store: MockStore, tower_jobs: list[dict], case_id: str):
    """Assess that jobs are successfully updated when using NF Tower."""

    # GIVEN an analysis without jobs
    analysis: Analysis | None = analysis_store.get_latest_analysis_for_case(case_id=case_id)

    # WHEN analysis jobs are deleted
    analysis_store.delete_analysis_jobs(analysis=analysis)

    # THEN the analysis object should have no jobs
    assert not analysis.jobs

    # WHEN jobs are updated
    analysis_store.update_jobs(jobs=tower_jobs[:2])

    # THEN jobs should be updated
    assert len(analysis.jobs) == 2


def test_update_analysis_with_comment(analysis_store: MockStore, case_id: str):
    # GIVEN an analysis without a comment
    analysis: Analysis = analysis_store.get_latest_analysis_for_case(case_id)
    analysis.comment = ""

    # WHEN updating the analysis with a comment
    analysis_store.update_analysis(analysis_id=analysis.id, comment="test")

    # THEN the comment should be set
    assert analysis.comment == "test"


def test_update_analysis_status(analysis_store: MockStore, case_id: str):
    # GIVEN an analysis with a non failed status
    analysis: Analysis = analysis_store.get_latest_analysis_for_case(case_id)
    assert analysis.status != "failed"

    # WHEN giving the analysis a status
    analysis_store.update_analysis(analysis_id=analysis.id, status="failed")

    # THEN the status should be set
    assert analysis.status == "failed"


def test_update_analysis_visibility(analysis_store: MockStore, case_id: str):
    # GIVEN an analysis which is not visible
    analysis: Analysis = analysis_store.get_latest_analysis_for_case(case_id)
    analysis.is_visible = False

    # WHEN making the analysis visible
    analysis_store.update_analysis(analysis_id=analysis.id, is_visible=True)

    # THEN the analysis should be visible
    assert analysis.is_visible



"""
Trailblazer is a tool to manage and track state of analyses.

You can install it from PyPI as follows::
    pip install trailblazer

For more detailed instructions, run :code:`trailblazer --help`

See the GitHub repo for documentation:
`Clinical-Genomics/analysis <https://github.com/Clinical-Genomics/analysis>`_
"""

import codecs
import io
import os

# Always prefer setuptools over distutils
from setuptools import find_packages, setup

AUTHOR = "henrik.stranneheim"
DESCRIPTION = "Trailblazer is a tool to manage and track state of analyses"
EMAIL = "henrik.stranneheim@scilifelab.se"
HERE = os.path.abspath(os.path.dirname(__file__))
NAME = "trailblazer"


def parse_reqs(req_path="./requirements.txt"):
    """Recursively parse requirements from nested pip files"""
    install_requires = []
    with codecs.open(req_path, "r") as handle:
        # remove comments and empty lines
        lines = (line.strip() for line in handle if line.strip() and not line.startswith("#"))
        for line in lines:
            # check for nested requirements files
            if line.startswith("-r"):
                # recursively call this function
                install_requires += parse_reqs(req_path=line[3:])
            else:
                # add the line as a new requirement
                install_requires.append(line)
    return install_requires


# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(HERE, "README.md"), encoding="utf-8") as f:
        LONG_DESCRIPTION = "\n" + f.read()
except FileNotFoundError:
    LONG_DESCRIPTION = DESCRIPTION


setup(
    name=NAME,
    version="21.5.36",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    keywords="analysis monitoring",
    author=AUTHOR,
    author_email=EMAIL,
    license="MIT",
    url="https://github.com/Clinical-Genomics/trailblazer",
    packages=find_packages(exclude=("tests*", "docs", "examples")),
    include_package_data=True,
    zip_safe=False,
    install_requires=parse_reqs(),
    tests_require=[
        "pytest",
    ],
    entry_points={
        "console_scripts": [
            "trailblazer = trailblazer.cli.core:base",
        ]
    },
    # See: http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.11",
        "Environment :: Console",
    ],
)



from pathlib import Path

from trailblazer.constants import FileFormat
from trailblazer.io.controller import ReadFile


def get_slurm_job_ids(job_id_file: str) -> list[int]:
    job_id_file_path = Path(job_id_file)
    content: dict = ReadFile.get_content_from_file(
        file_format=FileFormat.YAML, file_path=job_id_file_path
    )
    job_ids: list[int] = []
    for row in content.values():
        [job_ids.append(int(job_id)) for job_id in row]
    return job_ids



import datetime
import uuid

import pytest
from sqlalchemy.orm import Session

from trailblazer.constants import (
    PRIORITY_OPTIONS,
    TYPES,
    WORKFLOWS,
    JobType,
    TrailblazerStatus,
    WorkflowManager,
)
from trailblazer.store.database import get_session
from trailblazer.store.models import Analysis, Job


@pytest.fixture
def order_id_with_multiple_analyses() -> int:
    return 1


@pytest.fixture
def analysis(order_id_with_multiple_analyses) -> Analysis:
    analysis = Analysis(
        config_path="config_path",
        workflow="workflow",
        case_id="case_id",
        out_dir="out_dir",
        priority=PRIORITY_OPTIONS[0],
        started_at=datetime.datetime.now() - datetime.timedelta(weeks=1),
        status=TrailblazerStatus.PENDING,
        ticket_id="ticket_id",
        type=TYPES[0],
        workflow_manager=WorkflowManager.SLURM,
        is_visible=True,
        order_id=order_id_with_multiple_analyses,
    )
    session: Session = get_session()
    session.add(analysis)
    session.commit()
    session.add(analysis)
    session.commit()

    analysis_job_1 = Job(
        analysis_id=analysis.id,
        name="name",
        slurm_id=1,
        status=TrailblazerStatus.COMPLETED,
        started_at=datetime.datetime.now(),
        elapsed=1,
        job_type=JobType.ANALYSIS,
    )
    analysis_job_2 = Job(
        analysis_id=analysis.id,
        name="name",
        slurm_id=2,
        status=TrailblazerStatus.COMPLETED,
        started_at=datetime.datetime.now(),
        elapsed=1,
        job_type=JobType.ANALYSIS,
    )
    upload_job = Job(
        analysis_id=analysis.id,
        name="name",
        slurm_id=2,
        status=TrailblazerStatus.RUNNING,
        started_at=datetime.datetime.now(),
        elapsed=1,
        job_type=JobType.UPLOAD,
    )
    session.add_all([analysis_job_1, analysis_job_2, upload_job])
    session.commit()
    return analysis


@pytest.fixture
def analyses() -> list[Analysis]:
    """Analyses with different statuses, priorities, types, and workflows."""
    analyses: list[Analysis] = []
    for priority in PRIORITY_OPTIONS:
        for type in TYPES:
            for status in TrailblazerStatus.statuses():
                for workflow in WORKFLOWS:
                    analysis = Analysis(
                        config_path="config_path",
                        workflow=workflow,
                        case_id="case_id",
                        out_dir="out_dir",
                        priority=priority,
                        started_at=datetime.datetime.now(),
                        status=status,
                        ticket_id=str(uuid.uuid4()),
                        type=type,
                        workflow_manager=WorkflowManager.SLURM,
                        is_visible=True,
                        order_id=len(analyses),
                    )
                    analyses.append(analysis)
                analysis.comment = "comment"  # Ensure some analyses have a comment
    session: Session = get_session()
    session.add_all(analyses)
    return analyses


@pytest.fixture
def analysis_with_failed_job() -> Analysis:
    analysis = Analysis(
        config_path="config_path",
        workflow="workflow",
        case_id="case_id",
        out_dir="out_dir",
        priority=PRIORITY_OPTIONS[0],
        started_at=datetime.datetime.now(),
        status=TrailblazerStatus.FAILED,
        ticket_id="ticket_id",
        type=TYPES[0],
        workflow_manager=WorkflowManager.SLURM,
        is_visible=True,
    )

    failed_job = Job(
        analysis_id=analysis.id,
        name="name",
        slurm_id=1,
        status=TrailblazerStatus.FAILED,
        started_at=datetime.datetime.now(),
        elapsed=1,
    )
    running_job = Job(
        analysis_id=analysis.id,
        name="name",
        slurm_id=1,
        status=TrailblazerStatus.RUNNING,
        started_at=datetime.datetime.now(),
        elapsed=1,
    )
    analysis.jobs.append(failed_job)
    analysis.jobs.append(running_job)
    session = get_session()
    session.add_all([analysis, failed_job, running_job])
    session.commit()


@pytest.fixture
def non_existing_analysis_id() -> str:
    return "00"



from trailblazer.constants import TrailblazerStatus
from trailblazer.services.job_service import JobService
from trailblazer.store.models import Analysis


def test_job_progression_completed(job_service: JobService, analysis_with_completed_jobs: Analysis):
    # GIVEN an analysis with completed jobs

    # WHEN getting the analysis progression
    progression = job_service.get_analysis_progression(analysis_with_completed_jobs.id)

    # THEN the progression is 100%
    assert progression == 1.0


def test_job_progression_ongoing(job_service: JobService, analysis_with_running_jobs: Analysis):
    # GIVEN an analysis with ongoing jobs

    # WHEN getting the analysis progression
    progression = job_service.get_analysis_progression(analysis_with_running_jobs.id)

    # THEN the progression is less than 100%
    assert progression < 1.0


def test_job_progression_without_jobs(job_service: JobService, analysis_without_jobs: Analysis):
    # GIVEN an analysis without jobs

    # WHEN getting the analysis progression
    progression = job_service.get_analysis_progression(analysis_without_jobs.id)

    # THEN the progression is 0%
    assert progression == 0.0


def test_analysis_status_when_completed_jobs(
    job_service: JobService,
    analysis_with_completed_jobs: Analysis,
):
    # GIVEN an analysis with completed jobs

    # WHEN getting the analysis status
    status = job_service.get_analysis_status(analysis_with_completed_jobs.id)

    # THEN the status is completed
    assert status == TrailblazerStatus.COMPLETED


def test_analysis_status_when_running_jobs(
    job_service: JobService,
    analysis_with_running_jobs: Analysis,
):
    # GIVEN an analysis with running jobs

    # WHEN getting the analysis status
    status = job_service.get_analysis_status(analysis_with_running_jobs.id)

    # THEN the status is running
    assert status == TrailblazerStatus.RUNNING



from datetime import datetime

from pydantic import BaseModel, ConfigDict, Field, field_validator

from trailblazer.constants import (
    TOWER_PROCESS_STATUS,
    TOWER_TASK_STATUS,
    SlurmJobStatus,
    TrailblazerStatus,
)
from trailblazer.utils.datetime import tower_datetime_converter

SCALE_TO_MILLISEC: int = 1000


class TowerTask(BaseModel):
    """NF Tower task model."""

    process: str
    name: str
    status: str
    nativeId: str
    dateCreated: str | datetime | None = None
    lastUpdated: str | datetime | None = None
    start: str | datetime | None = None
    duration: int | None = None
    hash: str | None = None
    tag: str | None = None
    submit: str | None = None
    complete: str | None = None
    module: list[str]
    container: str | None = None
    attempt: int | None = None
    script: str | None = None
    scratch: str | None = None
    workdir: str | None = None
    queue: str | None = None
    cpus: int | None = None
    memory: int | None = None
    disk: int | None = None
    time: int | None = None
    env: str | None = None
    executor: str | None = None
    machineType: str | None = None
    cloudZone: str | None = None
    priceModel: str | None = None
    cost: float | None = None
    errorAction: str | None = None
    exitStatus: int | None = None
    realtime: int | None = None
    pcpu: float | None = None
    pmem: float | None = None
    rss: int | None = None
    vmem: int | None = None
    peakRss: int | None = None
    peakVmem: int | None = None
    rchar: int | None = None
    wchar: int | None = None
    syscr: int | None = None
    syscw: int | None = None
    readBytes: int | None = None
    writeBytes: int | None = None
    volCtxt: int | None = None
    invCtxt: int | None = None
    exit: str | None = None
    id: int | None = None
    taskId: int | None = None
    model_config = ConfigDict(validate_default=True)

    @field_validator("duration")
    @classmethod
    def set_duration(cls, raw_duration: int | None) -> int:
        """Convert milliseconds to seconds or return 0 if empty."""
        return round(raw_duration / SCALE_TO_MILLISEC) if raw_duration else 0

    @field_validator("status")
    @classmethod
    def set_status(cls, raw_status) -> str:
        return TOWER_TASK_STATUS.get(raw_status)

    @field_validator("start", "dateCreated", "lastUpdated")
    @classmethod
    def set_datetime(cls, raw_time) -> str | datetime | None:
        if isinstance(raw_time, str):
            return tower_datetime_converter(datetime_stamp=raw_time)
        elif isinstance(raw_time, datetime):
            return raw_time
        else:
            return None

    @property
    def is_complete(cls) -> bool:
        """Returns if the process succeded."""
        return cls.status == SlurmJobStatus.COMPLETED


class TaskWrapper(BaseModel):
    task: TowerTask


class TowerProcess(BaseModel):
    """NF Tower task model."""

    process: str
    dateCreated: str | datetime | None = None
    lastUpdated: str | datetime | None = None
    pending: int | None = None
    submitted: int | None = None
    running: int | None = None
    succeeded: int | None = None
    failed: int | None = None
    cached: int | None = None
    memoryEfficiency: float | None = None
    cpuEfficiency: float | None = None
    loadCpus: int | None = None
    loadMemory: int | None = None
    peakCpus: int | None = None
    peakTasks: int | None = None
    peakMemory: int | None = None
    model_config = ConfigDict(validate_default=True)

    @field_validator("dateCreated", "lastUpdated")
    @classmethod
    def set_datetime(cls, raw_time) -> str | datetime | None:
        if isinstance(raw_time, str):
            return tower_datetime_converter(datetime_stamp=raw_time)
        elif isinstance(raw_time, datetime):
            return raw_time
        else:
            return None

    @property
    def is_complete(cls) -> bool:
        """Returns if the process succeded."""
        return cls.status == TrailblazerStatus.COMPLETED

    @property
    def status(cls) -> str:
        for status_flag in TOWER_PROCESS_STATUS.keys():
            if cls.model_dump().get(status_flag, 0):
                return TOWER_PROCESS_STATUS.get(status_flag)
        return TrailblazerStatus.ERROR


class TowerWorkflow(BaseModel):
    """NF Tower workflow model."""

    status: str


class TowerProgress(BaseModel):
    """NF Tower progress model."""

    workflow_progress: dict = Field(alias="workflowProgress")
    processes_progress: list = Field(alias="processesProgress")


class TowerTasksResponse(BaseModel):
    """NF Tower task response model."""

    tasks: list[TaskWrapper] | None = None
    total: int

    def get_tasks(self) -> list[TowerTask]:
        return [task.task for task in self.tasks] if self.tasks else []


class TowerWorkflowResponse(BaseModel):
    """NF Tower task model."""

    workflow: TowerWorkflow
    progress: TowerProgress



import requests

from trailblazer.clients.tower.models import (
    TowerTasksResponse,
    TowerWorkflowResponse,
)
from trailblazer.clients.tower.utils import handle_client_errors


class TowerAPIClient:
    """A client consuming the Tower API. Endpoints are defined in https://tower.nf/openapi/."""

    def __init__(self, base_url: str, access_token: str, workspace_id: str):
        self.base_url = base_url
        self.request_params = [("workspaceId", workspace_id)]
        self.headers = {
            "Accept": "application/json",
            "Authorization": f"Bearer {access_token}",
        }

    @handle_client_errors
    def get_tasks(self, workflow_id: str) -> TowerTasksResponse:
        url = f"{self.base_url}workflow/{workflow_id}/tasks"
        response = requests.get(url=url, headers=self.headers, params=self.request_params)
        response.raise_for_status()
        json = response.json()
        return TowerTasksResponse.model_validate(json)

    @handle_client_errors
    def get_workflow(self, workflow_id: str) -> TowerWorkflowResponse:
        url = f"{self.base_url}workflow/{workflow_id}"
        response = requests.get(url=url, headers=self.headers, params=self.request_params)
        response.raise_for_status()
        json = response.json()
        return TowerWorkflowResponse.model_validate(json)

    @handle_client_errors
    def cancel_workflow(self, workflow_id: str) -> None:
        url = f"{self.base_url}workflow/{workflow_id}/cancel"
        response = requests.post(url=url, headers=self.headers, params=self.request_params, json={})
        response.raise_for_status()



from typing_extensions import Annotated
from pydantic import BaseModel, BeforeValidator, Field, ValidationInfo


def parse_order_ids(v: str) -> list[str]:
    return v[0].split(",") if v else []


class SummariesRequest(BaseModel):
    order_ids: Annotated[list[int], BeforeValidator(parse_order_ids)] = Field(alias="orderIds")



from enum import StrEnum
from pydantic import BaseModel, Field

from trailblazer.constants import (
    TrailblazerPriority,
    TrailblazerStatus,
    TrailblazerTypes,
)
from trailblazer.dto.common import SortOrder


class AnalysisSortField(StrEnum):
    CASE_ID: str = "case_id"
    TICKET_ID: str = "ticket_id"
    STARTED_AT: str = "started_at"
    STATUS: str = "status"
    UPLOADED_AT: str = "uploaded_at"


class AnalysesRequest(BaseModel):
    workflow: str = ""
    search: str | None = None
    page_size: int = Field(alias="pageSize", default=250)
    page: int = 1
    sort_field: AnalysisSortField = Field(alias="sortField", default=AnalysisSortField.STARTED_AT)
    sort_order: SortOrder = Field(alias="sortOrder", default=SortOrder.DESC)
    status: list[TrailblazerStatus] = []
    priority: list[TrailblazerPriority] = []
    type: list[TrailblazerTypes] = []
    has_comment: bool | None = Field(alias="hasComment", default=None)
    order_id: int | None = Field(alias="orderId", default=None)
    case_id: str | None = None
    delivered: bool | None = None
    include_hidden: bool | None = Field(alias="includeHidden", default=None)



from trailblazer.services.slurm.dtos import SlurmJobInfo
from trailblazer.store.models import Job


def create_job(job_dto: SlurmJobInfo) -> Job:
    return Job(
        slurm_id=job_dto.slurm_id,
        status=job_dto.status,
        name=job_dto.name,
        started_at=job_dto.started_at,
        elapsed=job_dto.elapsed,
    )



from datetime import datetime
from trailblazer.clients.slurm_api_client.dto.common import SlurmAPIJobInfo
from trailblazer.clients.slurm_api_client.dto.job_response import SlurmJobResponse
from trailblazer.constants import TrailblazerStatus
from trailblazer.services.slurm.dtos import SlurmJobInfo


def create_job_info_dto(job_response: SlurmJobResponse) -> SlurmJobInfo:
    job: SlurmAPIJobInfo = job_response.jobs[0]
    elapsed: int | None = get_job_elapsed_time(job)
    start_time: datetime | None = get_job_start_time(job)
    status: TrailblazerStatus | None = get_job_state(job)

    return SlurmJobInfo(
        slurm_id=job.job_id,
        status=status,
        name=job.name,
        started_at=start_time,
        elapsed=elapsed,
    )


def get_job_elapsed_time(job: SlurmAPIJobInfo) -> int:
    return job.accrue_time.number if job.accrue_time and job.accrue_time.set else 0


def get_job_start_time(job: SlurmAPIJobInfo) -> datetime | None:
    if job.start_time is None:
        return None
    return datetime.fromtimestamp(job.start_time.number)


def get_job_state(job: SlurmAPIJobInfo) -> TrailblazerStatus | None:
    return TrailblazerStatus[job.job_state[0]] if job.job_state else None



from trailblazer.models import Config


def test_instantiate_config():
    """Tests config input against a pydantic model."""
    # GIVEN config raw input

    # WHEN instantiating a Config object
    config = Config(database_url="a_database")

    # THEN it was successfully created
    assert isinstance(config, Config)
